# -*- coding: utf-8 -*-
#cython: embedsignature=True, language_level=3
#cython: boundscheck=False, wraparound=False, cdivision=True, initializedcheck=False,
#This is for developping
##cython: profile=True, warn.undeclared=True, warn.unused=True, warn.unused_result=False, warn.unused_arg=True
"""
Bayesian Inverse Fourier Transform

This code is the implementation of
Steen Hansen J. Appl. Cryst. (2000). 33, 1415-1421

Based on the BIFT from Jesse Hopkins, available at:
https://sourceforge.net/p/bioxtasraw/git/ci/master/tree/bioxtasraw/BIFT.py

This is a major rewrite in Cython
"""
cdef:
    list authors
    str __license__, __copyright__, __date__

__authors__ = ["Jerome Kieffer", "Jesse Hopkins"]
__license__ = "MIT"
__copyright__ = "2020, ESRF"
__date__ = "10/06/2020"

import time
import cython
from cython.parallel import prange
from cython.view cimport array as cvarray
import numpy
cimport numpy as cnumpy
from libc.math cimport sqrt, fabs, pi, sin, log, exp, isfinite

from scipy.linalg import lapack
from scipy.linalg.cython_lapack cimport dgesvd
from scipy.linalg.cython_blas cimport dgemm, ddot
import logging
import itertools
logger = logging.getLogger(__name__)

from .collections import RadiusKey, PriorKey, TransfoValue, EvidenceKey, EvidenceResult, StatsResult

################################################################################
# BLAS / LAPACK wrappers
################################################################################

cpdef inline double blas_ddot(double[::1] a, double[::1] b) nogil:
    "Wrapper for double precision dot product"
    cdef:
        int n, one=1
        double *a0=&a[0]
        double *b0=&b[0]

    n = a.shape[0]
    if n != b.shape[0]:
        with gil:
            raise ValueError("Shape mismatch in input arrays.")

    return ddot(&n, a0, &one, b0, &one)


cpdef int blas_dgemm(double[:,::1] a, double[:,::1] b, double[:,::1] c, double alpha=1.0, double beta=0.0) nogil except -1:
    "Wrapper for double matrix-matrix multiplication C = AxB "
    cdef:
        char *transa = 'n'
        char *transb = 'n'
        int m, n, k, lda, ldb, ldc
        double *a0=&a[0,0]
        double *b0=&b[0,0]
        double *c0=&c[0,0]

    ldb = (&a[1,0]) - a0 if a.shape[0] > 1 else 1
    lda = (&b[1,0]) - b0 if b.shape[0] > 1 else 1

    k = b.shape[0]
    if k != a.shape[1]:
        with gil:
            raise ValueError("Shape mismatch in input arrays.")
    m = b.shape[1]
    n = a.shape[0]
    if n != c.shape[0] or m != c.shape[1]:
        with gil:
            raise ValueError("Output array does not have the correct shape.")
    ldc = (&c[1,0]) - c0 if c.shape[0] > 1 else 1
    dgemm(transa, transb, &m, &n, &k, &alpha, b0, &lda, a0,
               &ldb, &beta, c0, &ldc)
    return 0


cpdef int lapack_svd(double[:, ::1] A, double[::1] eigen, double[::1] work) nogil except -1:
    cdef:
        char *jobN = 'n'
        int n, lda, lwork, info, one=1

    info = 0
    n = A.shape[0]
    lwork = work.shape[0]
    if n != A.shape[1]:
        return -1
    lda = (&A[1,0]) - &A[0,0] if A.shape[0] > 1 else 1
    dgesvd(jobN, jobN, &n, &n, &A[0,0], &lda, &eigen[0], &work[0] , &one, &work[0], &one, &work[0], &lwork, &info)
    if info:
        return -1
    return 0

################################################################################
# Helper functions
################################################################################

cpdef distribution_sphere(double I0,
                          double Dmax,
                          int npt):
    """Creates the initial P(r) function for the prior as a sphere.

    Formula from Svergun & Koch 2003:

    p(r) = 4*pi*Dmax**3/24*r**2*(1-1.5*r/Dmax+0.5*(r/Dmax)**3)

    Normalized so that the area is equal to the measured I(0). Integrate p(r):

    I(0) = integral of p(r) from 0 to Dmax
         = (4*pi*D^3/24)**2

    So the normalized p(r) is:

    p(r) = r**2*(1-1.5*r/Dmax+0.5*(r/Dmax)**3) * I0/(4*pi*Dmax**3/24)

    To match convention with old RAW BIFT, we also carry around an extra factor
    of 4*pi*Delta r, so the normalization becomes:

    p(r) = r**2*(1-1.5*r/Dmax+0.5*(r/Dmax)**3) * I0/(Dmax**3/(24*Delta_r))

    :param I0: forward scattering intensity
    :param npt: number of points in the real space (-1)
    :param Dmax: Diameter of the the object, here a sphere
    :return: the density p(r)
    """
    cdef:
        int j
        double norm, delta_r, r
        double[::1] p_r
    p_r = cvarray(shape=(npt+1,), itemsize=sizeof(double), format="d")
    with nogil:
        norm = I0 / (4.0 * pi * Dmax**3 / 24.0)
        delta_r = Dmax / npt
        for j in range(npt+1):
            r = j * delta_r
            p_r[j] = norm * r**2 * (1.0 - 1.5*(r/Dmax) + 0.5*(r/Dmax)**3)
    # p = p * I0/(4*pi*Dmax**3/24.)
    # p = p * I0/(Dmax**3/(24.*(r[1]-r[0])))   #Which normalization should I use? I'm not sure either agrees with what Hansen does.
    return numpy.asarray(p_r)


cpdef inline distribution_parabola(double I0,
                                   double Dmax,
                                   int npt):
    """Creates the initial p(r) function for the prior as a parabola.

    p(r) = r * (Dmax - r)

    I0 = 4*pi * Dmax³/6

    :param I0: forward scattering intensity
    :param npt: number of points in the real space (-1)
    :param Dmax: Diameter of the the object, here a sphere
    :return: the density p(r)
    """
    cdef:
        int j
        double norm, delta_r, r
        double[::1] p_r
    p_r = cvarray(shape=(npt+1,), itemsize=sizeof(double), format="d")

    with nogil:
        norm = I0 / (4.0 * pi * Dmax**3 / 6.0)
        delta_r = Dmax / npt
        for j in range(npt+1):
            r = j * delta_r
            p_r[j] = norm * r * (Dmax - r)
    return numpy.asarray(p_r)


cpdef inline double  calc_regularization(double[::1] p_r,
                                         double[::1] f_r,
                                         double[::1] sigma2) nogil:
    """Calculate the regularization factor as defined in eq. 19:

    regularization = numpy.sum((f[1:-1]-p[1:-1])**2/sigma2[1:-1])

    :param p_r: smoothed density
    :param f_r: prior density
    :param sigma2: variance of density, i.e. deviation on the density squared

    Nota: the first and last points are skipped as they are null by construction
    """
    cdef:
        int j, npt
        double tmp
    tmp = 0.0
    npt = p_r.shape[0] - 1
    for j in range(1, npt):
        tmp += (p_r[j] - f_r[j])**2 / sigma2[j]
    return tmp

cpdef inline double calc_rlogdet(double[::1] f_r,
                                 double[:, ::1] B,
                                 double alpha,
                                 double[:, ::1] U,
                                 double[::1] eigen,
                                 double[::1] work) nogil:
        """
        Calculate the log of the determinant of the the matrix U.
        This is part of the evidence.

        u = numpy.sqrt(numpy.abs(numpy.outer(f_r[1:-1], f_r[1:-1])))*B[1:-1, 1:-1]/alpha
        u[numpy.diag_indices(u.shape[0])] += 1
        w = numpy.linalg.svd(u, compute_uv = False)
        rlogdet = numpy.sum(numpy.log(numpy.abs(w)))

        :param f_r: density as function of r
        :param B: autocorrelation of the transformation matrix
        :param alpha: weight of the regularization
        :param npt: number of points (-1) for the density
        :param U: squarre matrix (npt-1, npt-1) for calculating the determinant
        :param eigen: vector with the eigenvalues of matrix U (size npt-1)
        :param work: some work-space buffer used by LAPACK for the SVD
        :return:  log(abs(det(U))) where
        """
        cdef:
            int j, k, npt
            double rlogdet
        npt = f_r.shape[0]-1
        for j in range(1, npt):
            for k in range(1, npt):
                U[j-1, k-1] = sqrt(fabs(f_r[j]*f_r[k]))*B[j, k]/alpha + (1.0 if j==k else 0.0)
        if lapack_svd(U, eigen, work):
            with gil:
                raise RuntimeError("SVD failed")
        rlogdet = 0.0
        for j in range(npt-1):
            rlogdet += log(fabs(eigen[j]))
        return rlogdet

cpdef inline void ensure_edges_zero(double[::1] distribution) nogil:
  """This function sets the first and last point of the density plot to 0
  :param distribution: raw density
  The operation is performed in place
  """
  npt = distribution.shape[0] - 1
  distribution[0] = 0.0
  distribution[npt] = 0.0

cpdef inline void smooth_density(double[::1] raw,
                                 double[::1] smooth) nogil:
        """This function applies the smoothing of the density plot

        :param raw: raw density, called *f* in eq.19
        :param smooth: smoothed density, called *m* in eq.19

        The smoothing is performed in place
        """
        cdef:
            int k, npt
        #assert raw.shape[0] == smooth.shape[0]
        npt = raw.shape[0] - 1
        #Set the edges to be equal
        smooth[0] = raw[0]
        smooth[npt] = raw[npt]
        for k in range(2, npt-1):
            smooth[k] = 0.5 * (raw[k-1] + raw[k+1])
        #Interpolate the second and second to last point
        smooth[1] = (smooth[0] + smooth[2]) * 0.5
        smooth[npt-1] = (smooth[npt-2] + smooth[npt]) * 0.5

################################################################################
# Main class
################################################################################

cdef class BIFT:
    """Bayesian Inverse Fourier Transform

    :param q: scattering vector in 1/nm or 1A, the unit of q imposes the one on Dmax, r, ...
    :param I: Scattering intensity I(q)
    :param I_std: error on the intensity estimation    """
    cdef:
        readonly int size, high_start, high_stop
        readonly double I0_guess, delta_q, Dmax_guess, alpha_max
        readonly double[::1] q, intensity, variance, wisdom
        readonly dict prior_cache, evidence_cache, radius_cache, transfo_cache, lapack_cache

    def __cinit__(self, q, I, I_std):
        """Constructor of the Cython class
        :param q: scattering vector in 1/nm or 1A, the unit of q imposes the one on Dmax, r, ...
        :param I: Scattering intensity I(q)
        :param I_std: error on the intensity estimation
        """
        self.size = q.shape[0]
        assert self.size == I.shape[0], "Intensity array matches in size"
        assert self.size == I_std.shape[0], "Error array matches in size"
        self.q = numpy.ascontiguousarray(q, dtype=numpy.float64)
        self.intensity = numpy.ascontiguousarray(I, dtype=numpy.float64)
        self.variance = numpy.ascontiguousarray(I_std**2, dtype=numpy.float64)
        self.delta_q = (q[self.size-1]-q[0]) / (q.size-1)
        self.wisdom = None
        #We define a region of high signal where the noise is expected to be minimal:
        self.I0_guess = numpy.max(I)  # might be replaced with replaced with data from the Guinier fit
        self.high_start = numpy.argmax(I) # Might be replaced by the guinier region
        self.high_stop = self.high_start + numpy.where(I[self.high_start:]<self.I0_guess/2.)[0][0]
        self.Dmax_guess = 0.0
        self.alpha_max = 0.0
        self.prior_cache = {}
        self.evidence_cache = {}
        self.radius_cache = {}
        self.transfo_cache = {}
        self.lapack_cache = {}


    def __dealloc__(self):
        "This is the destructor of the class: free the memory and empty all caches"
        self.reset()
        self.q = self.intensity = self.variance = None

    def reset(self):
        "rest all caches"
        cdef:
            dict cache
        for cache  in (self.prior_cache,  self.evidence_cache, self.radius_cache, self.transfo_cache, self.lapack_cache):
            if cache is not None:
                for key in list(cache.keys()):
                    cache.pop(key)

    def set_Guinier(self, guinier_fit, Dmax_over_Rg=3.0):
        """Set some starting point from Guinier fit like:

        Dmax = 3 Rg
        I0 for prior density modeling
        Guinier region for high signal with low noise section

        :param guinier_fit: RG_RESULT instance from autorg fit
        :param Dmax_over_Rg: guess the Dmax =  factor * Rg
        :return: guessed Dmax
        """
        self.I0_guess = guinier_fit.I0
        self.high_start = guinier_fit.start_point
        self.high_stop = guinier_fit.end_point
        self.Dmax_guess = guinier_fit.Rg * Dmax_over_Rg
        return self.Dmax_guess


    def guess_alpha_max(self, int npt):
        """This is to define the maximum realistic alpha value to scan for.

        idea: limit case is alpha·S0 = chi2/2

        so calculate the prior p, the regularization factor associated, ...

        :param npt: number of points in the real space (-1)
        """
        cdef:
            double[::1] density, smooth
            double[:, ::1] transfo
            double regularization, chi2
        if self.Dmax_guess<=0.0:
            raise RuntimeError("Please initialize with Guinier fit data using set_Guinier")
        density = self.prior_distribution(self.I0_guess, self.Dmax_guess, npt)
        ensure_edges_zero(density)
        smooth = numpy.zeros(npt+1, numpy.float64)
        smooth_density(density, smooth)
        regularization = calc_regularization(density, smooth, density) # eq19
        transfo = self.get_transformation_matrix(self.Dmax_guess, npt)
        chi2 = self.calc_chi2(transfo, density, npt)
        return 0.5*chi2/regularization

    def get_best(self):
        """Return the most probable configuration found so far and the number of valid
        """
        cdef:
            double best_evidence
            int nvalid
        best_evidence = numpy.finfo(numpy.float64).min
        best_key = None
        nvalid = 0
        for key, value in self.evidence_cache.items():
            if value.converged:
                nvalid += 1
                if value.evidence>best_evidence:
                    best_key = key
                    best_evidence = value.evidence
        if nvalid == 0:
            "None have converged yet ... take the least worse answer"
            for key, value in self.evidence_cache.items():
                if value.evidence>best_evidence:
                    best_key = key
                    best_evidence = value.evidence

        return best_key, self.evidence_cache.get(best_key), nvalid

    def prior_distribution(self,
                           double I0,
                           double Dmax,
                           int npt,
                           dist_type='sphere'):
        """Calculate the prior distribution for the bayesian analysis

        Implements memoizing.
        The memoising is performed with I0 = Dmax = 1. Rescaling is performd a posterori

        :param I0: forward scattering intensity, often approximated by th maximum intensity
        :param Dmax: Largest dimention of the object
        :param npt: number of points in the real space (-1)
        :param dist_type: Implements "sphere" and "parabola". "wisdom" is possible after initialization
        :return: the distance distribution function p(r) where r = numpy.linspace(0, Dmax, npt+1)

        Nota: the wisdom is the normalized best density found. It needs to be manually updated with update_wisdom()
        """
        key = PriorKey(dist_type, npt)
        if key in self.prior_cache:
            value = self.prior_cache[key]
        else:
            if dist_type == "sphere":
                value = self.prior_cache[key] = distribution_sphere(1, 1, npt)
            elif dist_type == "parabola":
                value = self.prior_cache[key] = distribution_parabola(1, 1, npt)
            else:
                raise RuntimeError("Only 'sphere' is accepted for dist_type")
        return (I0/Dmax) * value

    def update_wisdom(self):
        best_key, best_value, nvalid  =self.get_best()
        npt = best_key.npt
        if nvalid == 0:
            logger.warning("No converged solution was found. It is not advices to ")
            density = distribution_sphere(1, 1, npt)
        else:
            density = best_value.density / (4.0*pi*numpy.trapz(best_value.density, numpy.linspace(0, 1, npt+1)))
        key = PriorKey("wisdom", npt)
        self.prior_cache[key] = density

    def radius(self, Dmax, npt):
        """Calculate the radius array with memoizing

        Rationnal: the lookup in a dict is 1000x faster than numpy.linspace !

        :param Dmax: maximum distance/diameter
        :param npt: number of points in the real space (-1)

        :return: numpy array of size npt+1 from 0 to Dmax (both included)
        """
        key = RadiusKey(Dmax, npt)
        if key in self.radius_cache:
            value = self.radius_cache[key]
        else:
            value = self.radius_cache[key] = numpy.linspace(0, Dmax, npt+1)
        return value

    def get_transformation_matrix(self,
                                  double Dmax,
                                  int npt,
                                  bint all_=False):
        """Get the matrix T from the cache or calculates it:

        T.dot.p(r) = I(q)

        This is A_ij matrix in eq.2 of Hansen 2000.

        This function does the memoizing for T, B = T.T x T) and sum_dia

        :param Dmax: diameter or longest distance in the object
        :param npt: number of points in the real space (-1)
        :param all_: return in addition B and sum_dia arrays
        :return: the T matrix as: T.dot.p(r) = I(q)
        """
        cdef:
            double[::1] sum_dia
            double[:, ::1] B, transpo_mtx, transfo_mtx
        key = RadiusKey(Dmax, npt)
        if key in self.transfo_cache:
            value = self.transfo_cache[key]
        else:
            transfo_mtx  = cvarray(shape=(self.size, npt+1), itemsize=sizeof(double), format="d")
            transpo_mtx  = cvarray(shape=(npt+1, self.size), itemsize=sizeof(double), format="d")
            B = cvarray(shape=(npt+1, npt+1), itemsize=sizeof(double), format="d")
            sum_dia = cvarray(shape=(npt+1,), itemsize=sizeof(double), format="d")
            with nogil:
                self.initialize_arrays(Dmax, npt, transfo_mtx, transpo_mtx, B, sum_dia)
            value = self.transfo_cache[key] = TransfoValue(numpy.asarray(transfo_mtx), numpy.asarray(B), numpy.asarray(sum_dia))
        if all_:
            return value
        else:
            return value.transfo

    def get_workspace_size(self, npt):
        "This function calls LAPACK to measure the size of the workspace needed for the SVD"
        if npt in self.lapack_cache:
            value = self.lapack_cache[npt]
        else:
            _, s_lw = lapack.get_lapack_funcs(('gesvd', 'gesvd_lwork'))
            self.lapack_cache[npt] = value = lapack._compute_lwork(s_lw, npt, npt)
        return value

    cdef int initialize_arrays(self,
                            double Dmax,
                            int npt,
                            double[:, ::1] transf_matrix,
                            double[:, ::1] transp_matrix,
                            double[:, ::1] B,
                            double[::1] sum_dia
                            ) nogil except -1:

        cdef:
            double tmp, ql, prefactor, delta_r, il, varl
            int l, c, res

        delta_r = Dmax / npt
        prefactor = 4.0 * pi * delta_r
        sum_dia[:] = 0.0
        for l in range(self.size):
            ql = self.q[l] * delta_r
            il = self.intensity[l]
            varl = self.variance[l]
            for c in range(npt+1):
                tmp = ql * c
                tmp = prefactor * (sin(tmp)/tmp if tmp!=0.0 else 1.0)
                transf_matrix[l, c] = tmp
                sum_dia[c] += tmp * il / varl
                transp_matrix[c, l] = tmp / varl
        sum_dia[0] = 0.0

        res = blas_dgemm(transp_matrix, transf_matrix, B)
        if res:
            return -1
        #B = numpy.dot(TnT, T)
        B[0, :] = 0.0
        B[:, 0] = 0.0
        return 0

    def opti_evidence(self, param,
                      int npt, bint prior=0):
        """Function made for optimization based on the evidence maximisation

        :param parm: 2-tuple containing Dmax, log(alpha)
        :param npt: number of points in the real space (-1)
        :param prior: By default (False) use the sphere density to start with
                      Set to True to use the shape of the best found structure
        :return: -evidence for optimisation
        """
        cdef:
            double Dmax, logalpha
        Dmax, logalpha = param
        alpha = exp(logalpha)
        key = EvidenceKey(Dmax, alpha, npt)
        if key in self.evidence_cache:
            return -self.evidence_cache[key].evidence
        return -self.calc_evidence(Dmax, alpha, npt)

    cpdef double calc_evidence(self,
                               double Dmax,
                               double alpha,
                               int npt,
                               bint prior=False) with gil:
        """
        Calculate the evidence for the given set of parameters

        The Evidence is in Bayesian statistics log(P/(1-P)).
        According to eq.17, here evidence is log(P) only

        :param Dmax: diameter or longest distance in the object
        :param alpha: smoothing factor (>=0, not its log!)
        :param npt: number of points in the real space (-1)
        :param prior: By default (False) use the sphere density to start with
                      Set to True to use the shape of the best found structure
        :return: evidence (other results are cached)

        All the equation number are refering to
        J. Appl. Cryst. (2000). 33, 1415-1421
        """
        cdef:
            double[::1] radius, p_r, f_r, sigma2, sum_dia, workspace, eigen
            double[:, ::1] B, transfo_mtx, U
            double chi2, regularization, xprec, dotsp, rlogdet, evidence
            int j
            bint is_valid, converged

        xprec = 0.999
        key = EvidenceKey(Dmax, alpha, npt)
        #Simple checks: Dmax and alpha need to be positive
        if Dmax<=0:
            logger.info("Dmax negative: alpha=%s Dmax=%s", alpha, Dmax)
            self.evidence_cache[key] = EvidenceResult(-numpy.inf, numpy.NaN, numpy.NaN, numpy.NaN, numpy.NaN, False)
            return -numpy.inf
        if alpha<=0:
            logger.info("alpha negative: alpha=%s Dmax=%s", alpha, Dmax)
            self.evidence_cache[key] = EvidenceResult(-numpy.inf, numpy.NaN, numpy.NaN, numpy.NaN, numpy.NaN, False)
            return -numpy.inf

        # Here we perform all memory allocation for the complete function

        radius = self.radius(Dmax, npt)
        p_r = self.prior_distribution(self.I0_guess, Dmax, npt, dist_type="wisdom" if prior else "sphere")

        #here p_r is what Hansen calls m in eq.5
        p_r[0] = 0
        f_r = cvarray(shape=(npt+1,), itemsize=sizeof(double), format="d")
        #Note: f_r was called P in the original RAW BIFT code
        sigma2 = cvarray(shape=(npt+1,), itemsize=sizeof(double), format="d")

        #Those are used by the rlogdet function to calculate the determinant
        U = cvarray(shape=(npt-1, npt-1), itemsize=sizeof(double), format="d")
        eigen = cvarray(shape=(npt-1,), itemsize=sizeof(double), format="d")
        workspace = cvarray(shape=(self.get_workspace_size(npt-1),), itemsize=sizeof(double), format="d")

        transfo_mtx, B, sum_dia = self.get_transformation_matrix(Dmax, npt, all_=True)

        # At this stage, all buffers have been allocated ...
        with nogil:
            #Do some kind of rescaling of the input:
            #This would probably better be done on the large intensity region like Imax> I >Imax/2
            #c1 = numpy.sum(numpy.sum(transfo_mtx[1:4,1:-1]*p_r[1:-1], axis=1)/self.variance[1:4])
            #c2 = numpy.sum(numpy.asarray(self.intensity[1:4])/numpy.asarray(self.variance[1:4]))
            #print(c2/c1, self.scale_factor(transfo_mtx, p_r, 1, 4))
            self.scale_density(transfo_mtx, p_r, f_r, self.high_start, self.high_stop, npt, 1.001)

            # Do the optimization
            dotsp = self._bift_inner_loop(f_r, p_r, sigma2, B, alpha, npt, sum_dia, xprec=xprec)

            regularization = calc_regularization(p_r, f_r, sigma2) # eq19
            #chi2 =numpy.sum((numpy.asarray(self.intensity)[1:-1]-numpy.dot(transfo_mtx[1:-1,1:-1], (f_r)[1:-1]))**2/numpy.asarray(self.variance)[1:-1])/self.size
            chi2 = self.calc_chi2(transfo_mtx, f_r, npt) #  eq.6
            rlogdet = calc_rlogdet(f_r, B, alpha, U, eigen, workspace) # part of eq.20

            # The probablility is described in eq. 17, the evidence is apparently log(P)
            evidence = - log(Dmax) \
                       - alpha*regularization \
                       - 0.5 * chi2 \
                       - 0.5 * rlogdet \
                       - log(alpha)

            # Some kind of after the fact adjustment
            converged = (dotsp > xprec) # handles the case dotsp is Nan
            if not converged:
                # Make result much less likely, why 30 ?
                if evidence <= 0:
                    evidence *= 30.0
                else:
                    evidence /= 30.0

            #Check if those data are valid
            is_valid = isfinite(evidence)
            for j in range(npt+1):
                is_valid &= isfinite(f_r[j])
        # Store the results into the cache with the GIL
        if is_valid:
            self.evidence_cache[key] = EvidenceResult(evidence,
                                                      chi2/(self.size - npt),
                                                      regularization,
                                                      numpy.asarray(radius),
                                                      numpy.asarray(f_r),
                                                      converged)
            return evidence
        else:
            logger.info("Invalid evidence: Dmax: %s alpha: %s S: %s chi2: %s rlogdet:%s", Dmax, alpha, regularization, chi2, rlogdet)
            self.evidence_cache[key] = EvidenceResult(-numpy.inf, numpy.NaN, numpy.NaN, numpy.NaN, numpy.NaN, False)
            return -numpy.inf

    cdef double calc_chi2(self,
                           double[:, ::1] transfo,
                           double[::1] density,
                           int npt
                           )nogil:
        """Calculate chi²

        This is defined in eq.6

        chi² = sum[ (I(q) - Im(q))²/err(q)² ]

        where Im = T.dot.p(r)

        :param transfo: the tranformation matrix T
        :param density: the densty p(r)
        :return: chi²/size

        Former implementation:
        chi2 = numpy.sum((i[1:-1]-numpy.sum(T[1:-1,1:-1]*f[1:-1], axis=1))**2/err[1:-1])/i.size
        used to return the reduced chi², now the not reduced one
        """
        cdef:
            int idx_q, idx_r
            double chi2, Im

        chi2 = 0.0
        for idx_q in range(self.high_start, self.size):
            # Replace with dot-product
            Im = 0.0
            for idx_r in range(1, npt):
                Im += transfo[idx_q, idx_r] * density[idx_r]
            chi2 += ((Im - self.intensity[idx_q])**2/self.variance[idx_q])
        return chi2

    cdef double scale_density(self,
                             double[:, ::1] transfo,
                             double[::1] p_r,
                             double[::1] f_r,
                             int start,
                             int stop,
                             int npt,
                             float factor) nogil:
        """
        Do some kind of rescaling of the prior density

        This allows you to chose which q-regions to scale the data on.
        The original version says 1-4
        This would probably better be done on the large intensity region like Imax> I >Imax/2

        :param tranfo: the matrix which T.p(r) = I(q)
        :param density: the p(r) vector
        :param start: the start point in q-bin
        :param stop: the end point in q-bin
        :return: scale factor
        c1 = numpy.sum(numpy.sum(transfo[1:4,1:-1]*p_r[1:-1], axis=1)/self.variance[1:4])
        c2 = numpy.sum(numpy.asarray(self.intensity[1:4])/numpy.asarray(self.variance[1:4]))
        """
        cdef:
            int j
            double num, denom, tmp, v, scale_f, scale_p
        num = denom = 0.0
        for j in range(start, stop):
            v = self.variance[j]
            num += self.intensity[j] / v
            # Use a dot product
            tmp = blas_ddot(transfo[j, 1:npt], p_r[1:npt])
#             tmp = 0.0
#             for k in range(1, npt):
#                 tmp += transfo[j, k]*p_r[k]

            denom += tmp/v
#         with gil:
#             c1 = numpy.sum(numpy.sum(numpy.dot(transfo[1:4,1:-1],p_r[1:-1])/self.variance[1:4]))
#             c2 = numpy.sum(numpy.asarray(self.intensity[1:4])/numpy.asarray(self.variance[1:4]))
#             print(c2/c1, num/denom)
        scale_p = num/denom
        scale_f = factor * scale_p

        # Update data in place
        for j in range(1, npt):
            v = p_r[j]
            p_r[j] = v * scale_p
            f_r[j] = v * scale_f
        return num/denom

    cdef inline double _bift_inner_loop(self,
                                        double[::1] f_r,
                                        double[::1] p_r,
                                        double[::1] sigma2,
                                        double[:, ::1] B,
                                        double alpha,
                                        int npt,
                                        double[::1] sum_dia,
                                        double xprec) nogil:
        """
        Loop and seek for self consistence and smooth f

        Inner most function for BIFT ... most time is spent here.

        :param f_r: guess density of size N+1, limits are null. Both input and output array
        :param p_r: smoothed version of f over the 2 neighbours. Both input and output array
        :param sigma2: variance (std²) of p_r. Mainly an output array, assuming poissonnian
        :param B: 2D array (N+1 x N+1) with the autocorrelation of the transoformation matrix
        :param alpha: smoothing factor
        :param npt: number of points in the real space (-1)
        :param maxit: Maximum number of iteration: 2000
        :param minit: Minimum number of iteration: 200
        :param xprec: Precision expected (close to 1, but lower)
        :param omega: Upgrade dumping factor, [0=no update, 1=full update, default 0.5]
        :param epsilon: The minimum density (except at boundaries)
        :return: dotsp

        Nota: f_r, p_r and sigma2 are upgraded in place.
        """
        cdef:
            double dotsp, p_k, f_k, tmp_sum, fx, sigma_k, sum_dia_k, B_kk
            double sum_s2, sum_c2, sum_sc, s_k, c_k, omega, epsilon, denom
            int ite, k, maxit, minit
            bint is_valid

        #Define starting conditions
        maxit = 2000
        minit = 100
        omega = 0.5
        epsilon = 1e-10


        #loop variables
        dotsp = 0.0
        sum_c2 = sum_s2 = sum_sc = 0.0
        #Start loop
        for ite in range(maxit):
            if (ite > minit) and (dotsp > xprec):
                break
            is_valid = True
            #some kind of renormalization of the f & p vector to ensure positivity
            for k in range(1, npt):
                p_k = p_r[k]
                f_k = f_r[k]
                sigma2[k] = fabs(p_k + epsilon)
                if p_k <= 0:
                    p_r[k] = -p_k + epsilon
                if f_k <= 0:
                    f_r[k] = -f_k + epsilon

            #Set edges of f to zero
            ensure_edges_zero(f_r)
            #Apply smoothness constraint: p is the smoothed version of f
            smooth_density(f_r, p_r)

            #Calculate the next correction
            for k in range(1, npt):
                # a bunch of local variables:
                f_k = f_r[k]
                p_k = p_r[k]
                sigma_k = sigma2[k]
                sum_dia_k = sum_dia[k]
                B_kk = B[k, k]

                # fsumi = numpy.dot(B[k, 1:N], f[1:N]) - B_kk*f_k
#                 tmp_sum = 0.0
#                 for j in range(1, npt):
#                     tmp_sum += B[k, j] * f_r[j]
                tmp_sum = blas_ddot(B[k, 1:npt], f_r[1:npt])
                tmp_sum -= B[k, k]*f_r[k]

                fx = (2.0*alpha*p_k/sigma_k + sum_dia_k - tmp_sum) / (2.0*alpha/sigma_k + B_kk)
                is_valid &= isfinite(fx)
                # Finally update the value
                f_r[k] = f_k = (1.0-omega)*f_k + omega*fx
            #Synchro point

#             if not is_valid:
#                 with gil:
#                     for i in range(npt+1):
#                         print(i, f_r[i], p_r[i], sigma2[i])
#                 return 0.0

            # Calculate convergence
            sum_c2 = sum_sc = sum_s2 = 0.0
            for k in range(1, npt):
                s_k = 2.0 * (p_r[k] - f_r[k]) / sigma2[k] #Check homogeneity

#                 tmp_sum = 0.0
#                 for j in range(1, npt):
#                     tmp_sum += B[k, j] * f_r[j]
                tmp_sum = blas_ddot(B[k, 1:npt], f_r[1:npt])

                c_k = 2.0 * (tmp_sum - sum_dia[k])

                # There are 3 scalar products:
                sum_c2 += c_k * c_k
                sum_s2 += s_k * s_k
                sum_sc += s_k * c_k

            denom = sqrt(sum_s2*sum_c2)

    #         gradsi = 2.0*(p[1:N] - f[1:N])/sigma2[1:N]
    #         gradci = 2.0*(numpy.dot(B[1:N,1:N],f[1:N]) - sum_dia[1:N])

    #         wgrads2 = numpy.dot(gradsi,gradsi)
    #         wgradc2 = numpy.dot(gradci, gradci)
    #         denom = sqrt(wgrads2*wgradc2)
            if denom == 0.0:
                dotsp = 1.0
            else:
    #             dotsp = numpy.dot(gradsi,gradci) / denom
                dotsp = sum_sc / denom
        return dotsp

    def grid_scan(self,
                  double dmax_min, double dmax_max, int dmax_cnt,
                  double alpha_min, double alpha_max, int alpha_cnt,
                  int npt):
        """Perform a quick scan for Dmax and alpha

        :param alpha_min, alpha_max, alpha_cnt: parameters for a numpy.geomspace scan in alpha
        :param dmax_min, dmax_max, dmax_cnt: parameters for a numpy.linspace scan in Dmax
        :param npt: the number of points in the
        :return: best combination
        """
        cdef:
            double[:, ::1] grid
            double[::1] results, dmax_array, alpha_array
            int idx, best, steps = dmax_cnt*alpha_cnt,
            double Dmax, alpha
        dmax_array = numpy.linspace(dmax_min, dmax_max, dmax_cnt)
        alpha_array = numpy.geomspace(alpha_min, alpha_max, alpha_cnt)
        grid = numpy.array(list(itertools.product(dmax_array, alpha_array)))

        results = numpy.zeros(steps, dtype=numpy.float64)
#         for idx in range(steps):
        with nogil:
            for idx in prange(steps):
                Dmax = grid[idx, 0]
                alpha = grid[idx, 1]
                results[idx] += self.calc_evidence(Dmax, alpha, npt)
        best = numpy.argmax(results)
#         for i, j in zip(grid, results):
#             print(j, i[0], i[1], )
        return EvidenceKey(grid[best, 0], grid[best, 1], npt)

    def monte_carlo_sampling(self,
                             int samples,
                             double nsigma,
                             int npt
                             ):
        """Perform a monte-carlo sampling for Dmax and alpha around the optimum to be able to calculate the statistic of observables

        Dmax is sampled linearly
        alpha is sampled in log-scale

        Re-uses the results from the steepest descent to guess the std of alpha and Dmax.

        Re use the best density as prior for the density

        :param npt: the number of points in the modelisation
        :param samples: number of samples to be taken
        :param nsigma: sample alpha and Dmax at avg ± nx sigma
        :return: Statistics calculated over all explored space
        """
        cdef:
            double[::1] Dmax_samples, alpha_samples
            double[::1] results
            int idx
            double Dmax, alpha, t0, eps
        stats = self.calc_stats()
        if samples == 0:
            return stats
        self.update_wisdom()
        if stats.Dmax_std>0 and stats.Dmax_avg/stats.Dmax_std < nsigma:
            nsigma = stats.Dmax_avg/stats.Dmax_std
            logger.info("Clipping to nsigma=%.2f due to large noise on Dmax: avg=%.2f, std=%.2f", nsigma, stats.Dmax_avg, stats.Dmax_std)
        log_alpha = log(stats.alpha_avg)
        dlog_alpha = stats.alpha_std/stats.alpha_avg
        Dmax_samples = stats.Dmax_avg + nsigma*(2.0*numpy.random.random(samples)-1.0)*stats.Dmax_std
        alpha_samples = numpy.exp(log_alpha + nsigma*(2.0*numpy.random.random(samples)-1.0)*dlog_alpha)
        results = numpy.zeros(samples, dtype=numpy.float64)
        t0 = time.perf_counter()
        with nogil:
            for idx in prange(samples):
                Dmax = Dmax_samples[idx]
                alpha = alpha_samples[idx]
                results[idx] = self.calc_evidence(Dmax, alpha, npt, prior=1)
        logger.debug("Monte-carlo: %i samples at %.2fms/sample", samples, (time.perf_counter()-t0)*1000.0/samples)
        return self.calc_stats()

    def calc_stats(self):
        """Calculate the statistics on all points encountered so far ....

        :return: large namedtuple
        """
        cdef:
            int npt, nvalid, idx
            double area, ev_max, evidence_avg, evidence_std,
            double Dmax_avg, Dmax_std, alpha_avg, alpha_std, chi2_avg, chi2_std,
            double regularization_avg, regularization_std, Rg_std, Rg_avg, I0_avg, I0_std
            cnumpy.ndarray radius, densities, evidences, Dmaxs, alphas, chi2s, regularizations, proba, density_avg, density_std, areas, area2s, Rgs

        best_key, best, nvalid = self.get_best()
        if nvalid < 2:
            raise RuntimeError("Unable to calculate statistics without evidences having been optimized.")

        radius = best.radius
        npt = radius.size
        densities = numpy.zeros((nvalid, npt), dtype=numpy.float64)
        evidences = numpy.zeros(nvalid, dtype=numpy.float64)
        Dmaxs = numpy.zeros(nvalid, dtype=numpy.float64)
        alphas = numpy.zeros(nvalid, dtype=numpy.float64)
        chi2s = numpy.zeros(nvalid, dtype=numpy.float64)
        regularizations = numpy.zeros(nvalid, dtype=numpy.float64)

        idx = 0
        for key in self.evidence_cache:
            value = self.evidence_cache[key]
            if value.converged:
                Dmaxs[idx] = key.Dmax
                alphas[idx] = key.alpha
                evidences[idx] = value.evidence
                chi2s[idx] =  value.chi2r
                regularizations[idx] = value.regularization
                densities[idx] = numpy.interp(radius, value.radius, value.density, 0,0)
                idx+=1

        # Then, calculate the probability of each result as exp(evidence - evidence_max)**(1/minimum_chisq),
        # normalized by the sum of all result probabilities
        proba = numpy.exp(evidences - best.evidence)**(1./chi2s.min())
        proba /= proba.sum()

        #Then, calculate the average P(r) function as the weighted sum of the P(r) functions
        density_avg = numpy.sum(densities*proba[:,None], axis=0)
        #Then, calculate the error in P(r) as the square root of the weighted sum of squares of the difference between the average result and the individual estimate
        density_std = numpy.sqrt(numpy.abs(numpy.sum((densities-density_avg)**2*proba[:,None], axis=0)))

        #Then, calculate structural results as weighted sum of each result
        evidence_avg = numpy.dot(evidences, proba)
        evidence_std = numpy.sqrt(numpy.dot((evidences - evidence_avg)**2, proba))

        Dmax_avg = numpy.dot(Dmaxs, proba)
        Dmax_std = numpy.sqrt(numpy.dot((Dmaxs - Dmax_avg)**2, proba))

        alpha_avg = numpy.dot(alphas, proba)
        alpha_std = numpy.sqrt(numpy.dot((alphas - alpha_avg)**2, proba))

        chi2_avg = numpy.dot(chi2s, proba)
        chi2_std = numpy.sqrt(numpy.dot((chi2s - chi2_avg)**2, proba))

        regularization_avg = numpy.dot(regularizations, proba)
        regularization_std = numpy.sqrt(numpy.dot((regularizations - regularization_avg)**2, proba))

        areas = numpy.trapz(densities, radius, axis=1)
        area2s = numpy.trapz(densities*radius**2, radius, axis=1)

        Rgs = numpy.sqrt(area2s/(2.*areas))
        Rg_avg = numpy.sum(Rgs*proba)
        Rg_std = numpy.sqrt(numpy.sum((Rgs - Rg_avg)**2*proba))

        area = numpy.sum(areas*proba)
        I0_avg = 4.0 * pi * area
        I0_std = 4.0 * pi * numpy.sqrt(numpy.sum((areas - area)**2*proba))

        #Should I also extrapolate to q=0? Might be good, though maybe not in this function
        #Should I report number of good parameters (ftot(nmax-12 in Hansen code, line 2247))
        #Should I report number of Shannon Channels? That's easy to calculate: q_range*dmax/pi
        return StatsResult(radius, density_avg, density_std,
                           evidence_avg, evidence_std,
                           Dmax_avg, Dmax_std,
                           alpha_avg, alpha_std,
                           chi2_avg, chi2_std,
                           regularization_avg, regularization_std,
                           Rg_avg, Rg_std,
                           I0_avg, I0_std)



"""End to end tests for auto_gpa.py """

__authors__ = ["Martha Brennich"]
__license__ = "MIT"
__date__ = "16/07/2021"

import unittest
import pathlib
import logging
from subprocess import run, PIPE, STDOUT
from os import linesep
from os.path import normpath
from platform import system
from freesas.test.utilstests import get_datafile
logger = logging.getLogger(__name__)


class TestFreeSAS(unittest.TestCase):
    cwd = pathlib.Path.cwd()
    test_location = pathlib.Path(__file__)
    test_data_location = pathlib.Path(test_location.parent, "e2etest_data")
    bsa_filename = pathlib.Path(get_datafile("bsa_005_sub.dat"))
    sas_curve2_filename = pathlib.Path(get_datafile( "SASDF52.dat"))
    SASDFX7 = pathlib.Path(get_datafile("SASDFX7.dat"))

    def __init__(self, testName, **extra_kwargs):
        super().__init__(testName)
        self.extra_arg = extra_kwargs

    def test_one_bm29_bsa_without_arguments(self):
        """
        Test whether guinier fit apps on BSA data from BM29 returns one line.
        """
        app_name: str = self.extra_arg["app"]
        run_app = run(
            [app_name, normpath(str(self.bsa_filename))],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )
        self.assertEqual(
            run_app.returncode, 0, msg=f"{app_name} completed well"
        )
        if system() == "Windows":
            run_app_output = str(run_app.stdout)[:-1].replace("\\\\", "\\")
        else:
            run_app_output = str(run_app.stdout, "utf-8")[:-1]
        self.assertFalse(
            linesep in run_app_output,
            "One line of output of {app_name}",
        )
        self.assertTrue(
            str(self.bsa_filename) in run_app_output,
            msg="filename of testdata found in output of {app_name}",
        )
        self.assertTrue("I0" in run_app_output, msg=f"I0 in {app_name}")
        self.assertTrue("Rg" in run_app_output, msg=f"Rg in {app_name}")

    def test_two_easy_files_without_arguments(self):
        """
        Test whether auto_gpa.py with two datasets returns two lines.
        """
        app_name: str = self.extra_arg["app"]
        run_app = run(
            [
                app_name,
                str(self.bsa_filename),
                str(self.sas_curve2_filename),
            ],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )
        self.assertEqual(
            run_app.returncode,
            0,
            msg=f"{app_name} for 2 files completed well",
        )
        if system() == "Windows":
            run_app_output_raw = str(run_app.stdout)[:-1].replace("\\\\", "\\")
            run_app_output = run_app_output_raw.split("\\n")[:-1]
        else:
            run_app_output_raw = str(run_app.stdout, "utf-8")[:-1]
            run_app_output = run_app_output_raw.split(linesep)
        self.assertTrue(
            str(self.bsa_filename) in run_app_output_raw
            and str(self.sas_curve2_filename) in run_app_output_raw,
            f"filename of testdata found in output in output of {app_name}",
        )

        self.assertEqual(
            len(run_app_output),
            2,
            msg=f"Number of output lines is 2 in output of {app_name}",
        )
        for result in run_app_output:
            self.assertTrue(
                str(self.bsa_filename) in result
                or str(self.sas_curve2_filename) in result,
                f"filename of testdata not found in result line for {app_name}",
            )
            self.assertTrue("I" in result, f"I0 in {app_name} output")
            self.assertTrue("Rg" in result, f"Rg in {app_name} output")

    def test_csv_output_with_two_files(self):
        """Test whether the output in csv format is consistent"""
        app_name: str = self.extra_arg["app"]
        test_output_name = pathlib.Path(self.cwd, f"{app_name}.csv")
        try:
            test_output_name.unlink()
        except FileNotFoundError:
            pass
        run_app = run(
            [
                app_name,
                str(self.bsa_filename),
                str(self.sas_curve2_filename),
                "-o",
                str(test_output_name),
                "-f",
                "csv",
            ],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )
        self.assertEqual(
            run_app.returncode,
            0,
            msg=f"{app_name} for 2 files to csv completed well",
        )
        self.assertEqual(
            run_app.stdout,
            b"",
            msg=f"{app_name} for 2 files to csv provided no output to stdout",
        )
        with open(str(test_output_name), "r") as test_output_file:
            test_output_header = test_output_file.readline()[:-1].split(",")
            test_output_result = test_output_file.readlines()
        test_output_result = [
            line[:-1].split(",") for line in test_output_result
        ]
        expected_header = {
            "File",
            "Rg",
            "Rg StDev",
            "I(0)",
            "I(0) StDev",
            "First point",
            "Last point",
            "Quality",
            "Aggregated",
        }
        self.assertEqual(
            expected_header ^ set(test_output_header),
            set(),
            msg=f"header created by {app_name} contains the expected element",
        )
        self.assertEqual(
            len(test_output_header),
            len(set(test_output_header)),
            msg=f"header created by {app_name} contains no duplicates",
        )
        index = {}
        for header_item in expected_header:
            index[header_item] = test_output_header.index(header_item)

        self.assertEqual(
            len(test_output_result),
            2,
            msg=f"result file created by {app_name} contains 2 results",
        )
        self.assertEqual(
            {str(self.bsa_filename), str(self.sas_curve2_filename)},
            {
                test_output_result[0][index["File"]],
                test_output_result[1][index["File"]],
            },
            msg=f"result file created by {app_name} contains results for the right files",
        )
        for result in test_output_result:
            if result[index["File"]] == str(self.bsa_filename):
                self.assertTrue(
                    abs(float(result[index["Rg"]]) - 2.95) <= 0.1,
                    msg=f"Rg for BSA by {app_name} between 2.85 and 3.05",
                )
                self.assertTrue(
                    float(result[index["Rg StDev"]]) < 0.3,
                    msg=f"Rg std. for BSA by {app_name} below 0.3",
                )
                self.assertTrue(
                    abs(float(result[index["I(0)"]]) - 62.5) <= 2.5,
                    msg=f"I0 for BSA by {app_name} between 60 and 65",
                )
                self.assertTrue(
                    float(result[index["I(0) StDev"]]) < 5,
                    msg=f"I0 std. for BSA by {app_name} below 5",
                )
                self.assertTrue(
                    float(result[index["First point"]]) < 70,
                    msg=f"Guinier region start for BSA by {app_name} below 70",
                )
                self.assertTrue(
                    float(result[index["Last point"]]) > 85,
                    msg=f"Guinier region end for BSA by {app_name} above 85",
                )
                self.assertTrue(
                    float(result[index["Quality"]]) > 0.8,
                    msg=f"Fit quality BSA by {app_name} above 0.8",
                )
                self.assertTrue(
                    float(result[index["Aggregated"]]) < 0.05,
                    msg=f"Aggregation number for BSA by {app_name} below 0.05",
                )
            elif result[index["File"]] == str(self.sas_curve2_filename):
                self.assertTrue(
                    abs(float(result[index["Rg"]]) - 3.05) <= 0.2,
                    msg=f"Rg for BSA by {app_name} between 2.85 and 3.25",
                )
                self.assertTrue(
                    float(result[index["Rg StDev"]]) < 0.15,
                    msg=f"Rg std. for SASDF52 by {app_name} below 0.5",
                )
                self.assertTrue(
                    abs(float(result[index["I(0)"]]) - 64.5) <= 2.5,
                    msg=f"I0 for SASDF52 by {app_name} between 62 and 67",
                )
                self.assertTrue(
                    float(result[index["I(0) StDev"]]) < 5,
                    msg=f"I0 std. for SASDF52 by {app_name} below 5",
                )
                self.assertTrue(
                    float(result[index["First point"]]) < 35,
                    msg=f"Guinier region start for SASDF52 by {app_name} below 35",
                )
                self.assertTrue(
                    float(result[index["Last point"]]) > 80,
                    msg=f"Guinier region end for SASDF52 by {app_name} above 80",
                )
                self.assertTrue(
                    float(result[index["Quality"]]) > 0.8,
                    msg=f"Fit quality SASDF52 by {app_name} above 0.8",
                )
                self.assertTrue(
                    float(result[index["Aggregated"]]) < 0.1,
                    msg=f"Aggregation number for SASDF52 by {app_name} below 0.1",
                )

    def test_ssv_output_with_two_files(self):
        """Test wether the output in the ssv table has the correct dimensions"""
        app_name: str = self.extra_arg["app"]
        test_output_name = pathlib.Path(self.cwd, f"{app_name}.ssv")
        try:
            test_output_name.unlink()
        except FileNotFoundError:
            pass
        run_app = run(
            [
                app_name,
                str(self.bsa_filename),
                str(self.sas_curve2_filename),
                "-o",
                str(test_output_name),
                "-f",
                "ssv",
            ],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )
        self.assertEqual(
            run_app.returncode,
            0,
            msg=f"{app_name} for 2 files to csv completed well",
        )
        self.assertEqual(
            run_app.stdout,
            b"",
            msg=f"{app_name} for 2 files to ssv provided no output to stdout",
        )
        with open(str(test_output_name), "r") as test_output_file:
            test_output_result = test_output_file.readlines()
        test_output_result = [line[:-1].split() for line in test_output_result]

        self.assertEqual(
            len(test_output_result),
            2,
            msg=f"results created by {app_name} in native have the expected number of rows",
        )
        self.assertTrue(
            len(test_output_result[0]) == 9
            and len(test_output_result[1]) == 9,
            msg=f"results created by {app_name} in ssv have the expected number of elements",
        )
        self.assertEqual(
            {str(self.bsa_filename), str(self.sas_curve2_filename)},
            {
                test_output_result[0][-1],
                test_output_result[1][-1],
            },
            msg=f"ssv file created by {app_name} contains the filenames at the end of each row",
        )

    def test_native_output_with_two_files(self):
        """Test wether the output in the antive format has the correct dimensions"""
        app_name: str = self.extra_arg["app"]
        test_output_name = pathlib.Path(self.cwd, f"{app_name}.txt")
        try:
            test_output_name.unlink()
        except FileNotFoundError:
            pass
        run_app = run(
            [
                app_name,
                str(self.bsa_filename),
                str(self.sas_curve2_filename),
                "-o",
                str(test_output_name),
            ],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )
        self.assertEqual(
            run_app.returncode,
            0,
            msg=f"{app_name} for 2 files to native completed well",
        )
        self.assertEqual(
            run_app.stdout,
            b"",
            msg=f"{app_name} for 2 files to native provided no output to stdout",
        )
        with open(str(test_output_name), "r") as test_output_file:
            test_output_result = test_output_file.readlines()
        test_output_result = [line[:-1].split() for line in test_output_result]

        self.assertEqual(
            len(test_output_result),
            2,
            msg=f"results created by {app_name} in native have the expected number of rows",
        )
        self.assertTrue(
            len(test_output_result[0]) == 5
            and len(test_output_result[1]) == 5,
            msg=f"results created by {app_name} in native have the expected number of elements",
        )
        self.assertEqual(
            {str(self.bsa_filename), str(self.sas_curve2_filename)},
            {
                test_output_result[0][0],
                test_output_result[1][0],
            },
            msg=f"native file created by {app_name} contains "
            "the filenames at the start of each row",
        )

    def test_rg_of_SASDFX7_is_23Å(self):
        """
        SASDFX7 is an example of a dat file with artefacts at small angles.
        """
        app_name: str = self.extra_arg["app"]
        test_output_name = pathlib.Path(self.cwd, f"SASDFX7_{app_name}.csv")
        try:
            test_output_name.unlink()
        except FileNotFoundError:
            pass

        run_app = run(
            [
                app_name,
                str(self.SASDFX7),
                "-o",
                str(test_output_name),
                "-f",
                "csv",
                "-u",
                "Å",
            ],
            stdout=PIPE,
            stderr=STDOUT,
            check=True,
        )
        self.assertEqual(
            run_app.returncode,
            0,
            msg=f"{app_name} for SASDFX7 completed well",
        )

        with open(str(test_output_name), "r") as test_output_file:
            test_output_header = test_output_file.readline()[:-1].split(",")
            test_output_result = test_output_file.readline()
        result = test_output_result[:-1].split(",")

        expected_header = {
            "File",
            "Rg",
            "Rg StDev",
            "I(0)",
            "I(0) StDev",
            "First point",
            "Last point",
            "Quality",
            "Aggregated",
        }

        index = {}
        for header_item in expected_header:
            index[header_item] = test_output_header.index(header_item)

        self.assertTrue(
            abs(float(result[index["Rg"]]) - 2.3) <= 0.1,
            msg=f"Rg for SASDFX7 by {app_name} between 2.2 and 2.4",
        )


def suite():
    test_suite = unittest.TestSuite()
    for app in ["free_gpa", "free_guinier", "free_rg"]:
        if system() == "Windows":
            app += ".exe"
        test_suite.addTest(
            TestFreeSAS("test_one_bm29_bsa_without_arguments", app=app)
        )
        test_suite.addTest(
            TestFreeSAS("test_two_easy_files_without_arguments", app=app)
        )
        test_suite.addTest(
            TestFreeSAS("test_csv_output_with_two_files", app=app)
        )
        test_suite.addTest(
            TestFreeSAS("test_ssv_output_with_two_files", app=app)
        )
        test_suite.addTest(
            TestFreeSAS("test_native_output_with_two_files", app=app)
        )
        test_suite.addTest(TestFreeSAS("test_rg_of_SASDFX7_is_23Å", app=app))
    return test_suite


if __name__ == "__main__":
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python
# coding: utf-8
"""Print information about python."""

__authors__ = ["Jérôme Kieffer"]
__date__ = "08/12/2015"
__license__ = "MIT"


import sys

print("Python %s bits" % (tuple.__itemsize__ * 8))
print("       maxsize: %s\t maxunicode: %s" % (sys.maxsize, sys.maxunicode))
print(sys.version)
print(" ")

try:
    from distutils.sysconfig import get_config_vars
except ImportError:
    from sysconfig import get_config_vars
print("Config: " + str(get_config_vars("CONFIG_ARGS")))
print("")

try:
    import numpy
except ImportError:
    print("Numpy not installed")
else:
    print("Numpy %s" % numpy.version.version)
    print("      include %s" % numpy.get_include())
    print("      options %s" % numpy.get_printoptions())
print("")



#!/usr/bin/python3
# coding: utf-8
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2020  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "Jérôme Kieffer"
__license__ = "MIT"
__copyright__ = "2020, ESRF"
__date__ = "15/01/2021"

import io
import os
import sys
import logging
import glob
import platform
import posixpath
from collections import namedtuple, OrderedDict
import json
import copy
import pyFAI
from pyFAI.io import Nexus
from freesas.sas_argparser import SASParser

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("extract_ascii")

if sys.version_info[0] < 3:
    logger.error("This code requires Python 3.4+")

NexusJuice = namedtuple(
    "NexusJuice",
    "filename h5path npt unit q I poni mask energy polarization signal2d error2d buffer concentration",
)


def parse():

    """Parse input and return list of files.
    :return: list of input files
    """
    description = "Extract the SAXS data from a Nexus files as a 3 column ascii (q, I, err). Metadata are exported in the headers as needed."
    epilog = """extract_ascii.py allows you to export the data in inverse nm or inverse A with possible intensity scaling.
    """
    parser = SASParser(
        prog="extract-ascii.py", description=description, epilog=epilog
    )
    # Commented option need to be implemented
    # parser.add_argument("-o", "--output", action='store', help="Output filename, by default the same with .dat extension", default=None, type=str)
    # parser.add_argument("-u", "--unit", action='store', help="Unit for q: inverse nm or Angstrom?", default="nm", type=str)
    # parser.add_argument("-n", "--normalize", action='store', help="Re-normalize all intensities with this factor ", default=1.0, type=float)
    parser.add_file_argument("HDF5 input data")
    parser.add_argument(
        "-a",
        "--all",
        action="store_true",
        help="extract every individual frame",
        default=False,
    )
    return parser.parse_args()


def extract_averaged(filename):
    "return some infomations extracted from a HDF5 file "
    results = OrderedDict()
    results["filename"] = filename
    # Missing: comment normalization
    with Nexus(filename, "r") as nxsr:
        entry_grp = nxsr.get_entries()[0]
        results["h5path"] = entry_grp.name
        nxdata_grp = nxsr.h5[entry_grp.attrs["default"]]
        signal = nxdata_grp.attrs["signal"]
        axis = nxdata_grp.attrs["axes"]
        results["I"] = nxdata_grp[signal][()]
        results["q"] = nxdata_grp[axis][()]
        results["std"] = nxdata_grp["errors"][()]
        results["unit"] = pyFAI.units.to_unit(
            axis + "_" + nxdata_grp[axis].attrs["units"]
        )
        integration_grp = nxdata_grp.parent
        results["geometry"] = json.loads(
            integration_grp["configuration/data"][()]
        )
        results["polarization"] = integration_grp[
            "configuration/polarization_factor"
        ][()]

        instrument_grps = nxsr.get_class(entry_grp, class_type="NXinstrument")
        if instrument_grps:
            detector_grp = nxsr.get_class(
                instrument_grps[0], class_type="NXdetector"
            )[0]
            results["mask"] = detector_grp["pixel_mask"].attrs["filename"]
        sample_grp = nxsr.get_class(entry_grp, class_type="NXsample")[0]
        results["sample"] = posixpath.split(sample_grp.name)[-1]
        results["buffer"] = sample_grp["buffer"][()]
        results["storage temperature"] = sample_grp["temperature_env"][()]
        results["exposure temperature"] = sample_grp["temperature"][()]
        results["concentration"] = sample_grp["concentration"][()]
        if "2_correlation_mapping" in entry_grp:
            results["to_merge"] = entry_grp[
                "2_correlation_mapping/results/to_merge"
            ][()]
    return results


def extract_all(filename):
    "return some infomations extracted from a HDF5 file for  all individual frames"
    res = []
    results = OrderedDict()
    results["filename"] = filename
    with Nexus(filename, "r") as nxsr:
        entry_grp = nxsr.get_entries()[0]
        results["h5path"] = entry_grp.name
        nxdata_grp = nxsr.h5[entry_grp.name + "/1_integration/results"]
        signal = nxdata_grp.attrs["signal"]
        axis = nxdata_grp.attrs["axes"][1]
        I = nxdata_grp[signal][()]
        results["q"] = nxdata_grp[axis][()]
        std = nxdata_grp["errors"][()]
        results["unit"] = pyFAI.units.to_unit(
            axis + "_" + nxdata_grp[axis].attrs["units"]
        )
        integration_grp = nxdata_grp.parent
        results["geometry"] = json.loads(
            integration_grp["configuration/data"][()]
        )
        results["polarization"] = integration_grp[
            "configuration/polarization_factor"
        ][()]
        instrument_grp = nxsr.get_class(entry_grp, class_type="NXinstrument")[
            0
        ]
        detector_grp = nxsr.get_class(instrument_grp, class_type="NXdetector")[
            0
        ]
        results["mask"] = detector_grp["pixel_mask"].attrs["filename"]
        sample_grp = nxsr.get_class(entry_grp, class_type="NXsample")[0]
        results["sample"] = posixpath.split(sample_grp.name)[-1]
        results["buffer"] = sample_grp["buffer"][()]
        if "temperature_env" in sample_grp:
            results["storage temperature"] = sample_grp["temperature_env"][()]
        if "temperature" in sample_grp:
            results["exposure temperature"] = sample_grp["temperature"][()]
        if "concentration" in sample_grp:
            results["concentration"] = sample_grp["concentration"][()]
    #         if "2_correlation_mapping" in entry_grp:
    #             results["to_merge"] = entry_grp["2_correlation_mapping/results/to_merge"][()]
    for i, s in zip(I, std):
        r = copy.copy(results)
        r["I"] = i
        r["std"] = s
        res.append(r)
    return res


def write_ascii(results, output=None, hdr="#", linesep=os.linesep):
    """
        :param resusts: dict containing some NexusJuice
        :param output: name of the 3-column ascii file to be written
        :param hdr: header mark, usually '#'
        :param linesep: to be able to addapt the end of lines

    Adam Round explicitelly asked for (email from Date: Tue, 04 Oct 2011 15:22:29 +0200) :
    Modification from:
    # BSA buffer
    # Sample c= 0.0 mg/ml (these two lines are required for current DOS pipeline and can be cleaned up once we use EDNA to get to ab-initio models)
    #
    # Sample environment:
    # Detector = Pilatus 1M
    # PixelSize_1 = 0.000172
    # PixelSize_2 = 6.283185 (I think it could avoid confusion if we give teh actual pixel size as 0.000172 for X and Y and not to give the integrated sizes. Also could there also be a modification for PixelSize_1 as on the diagonal wont it be the hypotenuse (0.000243)? and thus will be on average a bit bigger than 0.000172)
    #
    # title = BSA buffer
    # Frame 7 of 10
    # Time per frame (s) = 10
    # SampleDistance = 2.43
    # WaveLength = 9.31e-11
    # Normalization = 0.0004885
    # History-1 = saxs_angle +pass -omod n -rsys normal -da 360_deg -odim = 1 /data/id14eh3/inhouse/saxs_pilatus/Adam/EDNAtests/2d/dumdum_008_07.edf/data/id14eh3/inhouse/saxs_pilatus/Adam/EDNAtests/misc/dumdum_008_07.ang
    # DiodeCurr = 0.0001592934
    # MachCurr = 163.3938
    # Mask = /data/id14eh3/archive/CALIBRATION/MASK/Pcon_01Jun_msk.edf
    # SaxsDataVersion = 2.40
    #
    # N 3
    # L q*nm  I_BSA buffer  stddev
    #
    # Sample Information:
    # Storage Temperature (degrees C): 4
    # Measurement Temperature (degrees C): 10
    # Concentration: 0.0
    # Code: BSA
    s-vector Intensity Error
    s-vector Intensity Error
    s-vector Intensity Error
    s-vector Intensity Error
    """
    hdr = str(hdr)
    headers = []
    if "comments" in results:
        headers.append(hdr + " " + results["comments"])
    else:
        headers.append(hdr)
    headers.append(
        hdr + " Sample c= %s mg/ml" % results.get("concentration", -1)
    )
    headers += [hdr, hdr + " Sample environment:"]
    if "geometry" in results:
        headers.append(
            hdr + " Detector = %s" % results["geometry"]["detector"]
        )
        headers.append(
            hdr + " SampleDistance = %s" % results["geometry"]["dist"]
        )
        headers.append(
            hdr + " WaveLength = %s" % results["geometry"]["wavelength"]
        )
    headers.append(hdr)
    if "comments" in results:
        headers.append(hdr + " title = %s" % results["comment"])
    if "to_merge" in results:
        headers.append(
            hdr
            + " Frames merged: "
            + " ".join([str(i) for i in results["to_merge"]])
        )
    if "normalization" in results:
        headers.append(hdr + " Normalization = %s" % results["normalization"])
    if "mask" in results:
        headers.append(hdr + " Mask = %s" % results["mask"])
    headers.append(hdr)
    headers.append(hdr + (" N 3" if "std" in results else " N 2"))
    line = hdr + " L "
    if "unit" in results:
        a, b = str(results["unit"]).split("_")
        line += a + "*" + b.strip("^-1") + "  I_"
    else:
        line += "q  I_"
    if "comment" in results:
        line += results["comments"]
    if "std" in results:
        line += "  stddev"
    headers.append(line)
    headers.append(hdr)
    headers.append(hdr + " Sample Information:")
    if "storage temperature" in results:
        headers.append(
            hdr
            + " Storage Temperature (degrees C): %s"
            % results["storage temperature"]
        )
    if "exposure temperature" in results:
        headers.append(
            hdr
            + " Measurement Temperature (degrees C): %s"
            % results["exposure temperature"]
        )

    headers.append(
        hdr + " Concentration: %s" % results.get("concentration", -1)
    )
    if "buffer" in results:
        headers.append(hdr + " Buffer: %s" % results["buffer"])
    headers.append(hdr + " Code: %s" % results.get("sample", ""))

    def write(headers, file_):

        file_.writelines(linesep.join(headers))
        file_.write(linesep)

        if "std" in results:
            data = [
                "%14.6e\t%14.6e\t%14.6e" % (q, I, std)
                for q, I, std in zip(
                    results["q"], results["I"], results["std"]
                )
            ]
        else:
            data = [
                "%14.6e\t%14.6e\t" % (q, I)
                for q, I in zip(results["q"], results["I"])
            ]
        data.append("")
        file_.writelines(linesep.join(data))

    if output:
        with open(output, "w") as f:
            write(headers, f)
    else:
        f = io.StringIO()
        write(headers, f)
        f.seek(0)
        return f.read()


def main():
    args = parse()
    if args.verbose:
        logging.root.setLevel(logging.DEBUG)
    files = [i for i in args.file if os.path.exists(i)]
    if platform.system() == "Windows" and files == []:
        files = glob.glob(args.file[0])
        files.sort()
    input_len = len(files)
    logger.debug("%s input files", input_len)
    for src in files:
        if args.all:
            dest = os.path.splitext(src)[0] + "%04i.dat"
            for idx, frame in enumerate(extract_all(src)):
                print(src, " --> ", dest % idx)
                write_ascii(frame, dest % idx)
        else:
            dest = os.path.splitext(src)[0] + ".dat"
            write_ascii(extract_averaged(src), dest)
            print(src, " --> ", dest)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding: utf-8
from __future__ import print_function

__author__ = "Guillaume"
__license__ = "MIT"
__copyright__ = "2015, ESRF"

import numpy
import unittest
import os
import tempfile
from .utilstests import get_datafile
from ..model import SASModel
from ..transformations import translation_from_matrix, euler_from_matrix
import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("SASModel_test")


def assign_random_mol(inf=None, sup=None):
    if not inf:
        inf = 0
    if not sup:
        sup = 100
    molecule = numpy.random.randint(inf, sup, size=400).reshape(100, 4).astype(float)
    molecule[:, -1] = 1.0
    m = SASModel(molecule)
    return m


class TesttParser(unittest.TestCase):
    testfile = get_datafile("model-01.pdb")

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.tmpdir = tempfile.mkdtemp()
        self.outfile = os.path.join(self.tmpdir, "out.pdb")

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        for fn in (self.outfile, self.tmpdir):
            if os.path.exists(fn):
                if os.path.isdir(fn):
                    os.rmdir(fn)
                else:
                    os.unlink(fn)

    def test_same(self):
        m = SASModel()
        m.read(self.testfile)
        m.save(self.outfile)
        infile = open(self.testfile).read()
        outfile = open(self.outfile).read()
        self.assertEqual(infile, outfile, msg="file content is the same")

    def test_rfactor(self):
        m = SASModel()
        m.read(self.testfile)
        n = SASModel()
        n.read(self.testfile)
        self.assertEqual(m.rfactor, n.rfactor, msg="R-factor is not the same %s != %s" % (m.rfactor, n.rfactor))

    def test_init(self):
        m = SASModel()
        m.read(self.testfile)
        n = SASModel(self.testfile)
        param1 = m.rfactor
        param2 = n.rfactor
        self.assertEqual(param1, param2, msg="pdb file not read correctly")

    def test_centroid(self):
        m = assign_random_mol()
        m.centroid()
        if len(m.com) != 3:
            logger.error("center of mass has not been saved correctly : length of COM position vector = %s!=3" % (len(m.com)))
        mol_centered = m.atoms[:, 0:3] - m.com
        center = mol_centered.mean(axis=0)
        norm = (center * center).sum()
        self.assertAlmostEqual(norm, 0, 12, msg="molecule is not centered : norm of the COM position vector %s!=0" % (norm))

    def test_inertia_tensor(self):
        m = assign_random_mol()
        m.inertiatensor()
        tensor = m.inertensor
        assert tensor.shape == (3, 3), "inertia tensor has not been saved correctly : shape of inertia matrix = %s" % (tensor.shape)

    def test_canonical_translate(self):
        m = assign_random_mol()
        trans = m.canonical_translate()
        if trans.shape != (4, 4):
            logger.error("pb with translation matrix shape: shape=%s" % (trans.shape))
        com = m.com
        com_componants = [com[0], com[1], com[2]]
        trans_vect = [-trans[0, -1], -trans[1, -1], -trans[2, -1]]
        self.assertEqual(com_componants, trans_vect, msg="do not translate on canonical position")

    def test_canonical_rotate(self):
        m = assign_random_mol()
        rot = m.canonical_rotate()
        if rot.shape != (4, 4):
            logger.error("pb with rotation matrix shape")
        if not m.enantiomer:
            logger.error("enantiomer has not been selected")
        det = numpy.linalg.det(rot)
        self.assertAlmostEqual(det, 1, 10, msg="rotation matrix determinant is not 1: %s" % (det))

    def test_canonical_parameters(self):
        m = assign_random_mol()
        m.canonical_parameters()
        can_param = m.can_param
        if len(can_param) != 6:
            logger.error("canonical parameters has not been saved properly")
        com_trans = translation_from_matrix(m.canonical_translate())
        euler_rot = euler_from_matrix(m.canonical_rotate())
        out_param = [com_trans[0], com_trans[1], com_trans[2], euler_rot[0], euler_rot[1], euler_rot[2]]
        self.assertEqual(can_param, out_param, msg="canonical parameters are not the good ones")

    def test_dist(self):
        m = assign_random_mol()
        n = SASModel(m.atoms)
        distance = m.dist(n, m.atoms, n.atoms)
        self.assertEqual(distance, 0, msg="NSD different of 0: %s!=0" % (distance))

    def test_can_transform(self):
        m = assign_random_mol()
        m.canonical_parameters()
        p0 = m.can_param
        mol1 = m.transform(p0, [1, 1, 1])
        if abs(mol1 - m.atoms).max() == 0:
            logger.error("molecule did not move")
        m.atoms = mol1
        m.centroid()
        m.inertiatensor()
        com = m.com
        tensor = m.inertensor
        diag = numpy.eye(3)
        matrix = tensor - tensor * diag
        self.assertAlmostEqual(abs(com).sum(), 0, 10, msg="molecule not on its center of mass")
        self.assertAlmostEqual(abs(matrix).sum(), 0, 10, "inertia moments unaligned ")

    def test_dist_move(self):
        m = assign_random_mol()
        n = SASModel(m.atoms)
        m.canonical_parameters()
        n.canonical_parameters()
        if abs(n.atoms - m.atoms).max() != 0:
            logger.error("molecules are different")
        p0 = m.can_param
        dist_after_mvt = m.dist_after_movement(p0, n, [1, 1, 1])
        self.assertEqual(dist_after_mvt, 0, msg="NSD different of 0: %s!=0" % (dist_after_mvt))

    def test_reverse_transform(self):
        m = assign_random_mol()
        n = SASModel(m.atoms)
        m.canonical_parameters()
        m.atoms = m.transform(m.can_param, [1, 1, 1], reverse=None)
        m.atoms = m.transform(m.can_param, [1, 1, 1], reverse=True)
        dist = m.dist(n, m.atoms, n.atoms)
        self.assertAlmostEqual(dist, 0.0, 10, msg="pb with reverse transformation : %s != 0.0" % dist)


def suite():
    testSuite = unittest.TestSuite()
    testSuite.addTest(TesttParser("test_same"))
    testSuite.addTest(TesttParser("test_rfactor"))
    testSuite.addTest(TesttParser("test_init"))
    testSuite.addTest(TesttParser("test_centroid"))
    testSuite.addTest(TesttParser("test_inertia_tensor"))
    testSuite.addTest(TesttParser("test_canonical_translate"))
    testSuite.addTest(TesttParser("test_canonical_rotate"))
    testSuite.addTest(TesttParser("test_canonical_parameters"))
    testSuite.addTest(TesttParser("test_dist"))
    testSuite.addTest(TesttParser("test_can_transform"))
    testSuite.addTest(TesttParser("test_dist_move"))
    testSuite.addTest(TesttParser("test_reverse_transform"))
    return testSuite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



#Cython module to calculate distances of set of atoms

__author__ = "Jerome Kieffer"
__license__ = "MIT"
__copyright__ = "2015, ESRF"

import sys
import cython
cimport numpy
import numpy
from cython cimport floating
from libc.math cimport sqrt, fabs, exp


@cython.wraparound(False)
@cython.boundscheck(False)
def calc_invariants(floating[:, :] atoms):
    """
    Calculate the invariants of the structure, i.e fineness, radius of gyration and diameter of the model.

    Nota: to economize size*numpy.sqrt, the sqrt is taken at the end of the calculation. 
    We should have done s += sqrt(d) and then s/size, but we do s+= d and then sqrt(s/size).
    You can check that the result is the same.

    @param atoms: 2d-array with atom coordinates:[[x,y,z],...]
    @return: 3-tuple containing (fineness, Rg, Dmax)
        * average distance between an atoms and its nearest neighbor
        * radius of gyration of the model
        * diameter of the model
    """
    cdef:
        int i, j, size
        floating d, x1, y1, z1, dx, dy, dz, big, d2, sum_d2, d2max
    size = atoms.shape[0]
    assert atoms.shape[1] >= 3
    big = sys.maxsize
    s = 0.0
    sum_d2 = 0.0 
    d2max = 0.0
    for i in range(size):
        x1 = atoms[i, 0]
        y1 = atoms[i, 1]
        z1 = atoms[i, 2]
        d = big
        for j in range(size):
            if i == j:
                continue
            dx = atoms[j, 0] - x1
            dy = atoms[j, 1] - y1
            dz = atoms[j, 2] - z1
            d2 = dx * dx + dy * dy + dz * dz
            sum_d2 += d2
            d2max = max(d2max, d2)
            d = min(d, d2)
        s += d
    return sqrt(s / size), sqrt(sum_d2 / 2.0) / size, sqrt(d2max)


cdef inline floating hard_sphere(floating pos, floating radius)nogil:
    """Density using hard spheres
    @param pos: fabs(d1-d)
    """
    if pos > 2.0 * radius:
        return 0.0 
    return (4 * radius + pos) * (2 * radius - pos) ** 2 / (16.0 * radius ** 3)

cdef inline floating soft_sphere(floating pos, floating radius)nogil:
    """Density using soft spheres (gaussian density)
    @param pos: fabs(d1-d)
    @param radius: radius of the equivalent hard sphere
    """
    cdef floating sigma = 0.40567 * radius
    return exp(- pos * pos / (2.0 * sigma * sigma)) * 0.3989422804014327 / sigma


IF HAVE_OPENMP:
    include "_distance_omp.pxi"
ELSE:
    include "_distance_nomp.pxi"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#    Project: FreeSAS
#             https://github.com/kif/freesas
#
#    Copyright (C) European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#

"""
Contains helper functions for loading SAS data from differents sources.
"""
__authors__ = ["Martha Brennich"]
__contact__ = "martha.brennich@googlemail.com"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "25/07/2020"
__status__ = "development"
__docformat__ = "restructuredtext"

from typing import List, Union
from os import PathLike
from numpy import loadtxt, array, ndarray

PathType = Union[PathLike, str, bytes]


def load_scattering_data(filename: PathType) -> ndarray:
    """
    Load scattering data q, I, err into a numpy array.

    :param filename: ASCII file, 3 column (q,I,err)
    :return: numpy array with 3 column (q,I,err)
    """
    try:
        data = loadtxt(filename)
    except OSError:
        raise OSError("File could not be read.")
    except ValueError:
        try:
            with open(filename) as data_file:
                text = data_file.readlines()
        except OSError:
            raise OSError("File could not be read.")
        else:
            try:
                data = parse_ascii_data(text, number_of_columns=3)
            except ValueError:
                raise ValueError(
                    "File does not seem to be " "in the format q, I, err. "
                )
    return data


def parse_ascii_data(
    input_file_text: List[str], number_of_columns: int
) -> ndarray:
    """
    Parse data from an ascii file into an N column numpy array

    :param input_file_text: List containing one line of input data per element
    :param number_of_columns: Expected number of lines in the data file
    :return: numpy array with 3 column (q,I,err)
    """
    data = []
    for line in input_file_text:
        split = line.split()
        if len(split) == number_of_columns:
            try:
                data.append([float(x) for x in split])
            except ValueError as err:
                if "could not convert string to float" in err.args[0]:
                    pass
                else:
                    raise
    if data == []:
        raise ValueError
    return array(data)


def convert_inverse_angstrom_to_nanometer(
    data_in_inverse_angstrom: ndarray,
) -> ndarray:
    """
    Convert data with q in 1/Å to 1/nm.

    :param data_in_inverse_angstrom: numpy array in format
                                     (q_in_inverse_Angstrom,I,err)
    :return: numpy array with 3 column (q_in_inverse_nm,I,err)
    """
    q_in_angstrom, intensity, err = data_in_inverse_angstrom.T
    return array([q_in_angstrom * 10.0, intensity, err]).T



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016-2018 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ############################################################################*/

__authors__ = ["V. Valls"]
__license__ = "MIT"
__date__ = "12/05/2020"

from numpy.distutils.misc_util import Configuration
import platform
import os
import numpy


def create_extension_config(name, extra_sources=None, can_use_openmp=False):
    """
    Util function to create numpy extension from the current pyFAI project.
    Prefer using numpy add_extension without it.
    """
    include_dirs = [numpy.get_include()]

    if can_use_openmp:
        extra_link_args = ['-fopenmp']
        extra_compile_args = ['-fopenmp']
    else:
        extra_link_args = []
        extra_compile_args = ["-ftree-vectorize"]

    sources = ["%s.pyx" % name]
    if extra_sources is not None:
        sources.extend(extra_sources)

    config = dict(
        name=name,
        sources=sources,
        include_dirs=include_dirs,
        language='c',
        extra_link_args=extra_link_args,
        extra_compile_args=extra_compile_args
    )

    return config


def configuration(parent_package='', top_path=None):
    config = Configuration('freesas', parent_package, top_path)

    ext_modules = [
        create_extension_config("_distance", can_use_openmp=True),
        create_extension_config("_cormap", can_use_openmp=False),
        create_extension_config("_autorg", can_use_openmp=False),
        create_extension_config("_bift", can_use_openmp=False),
    ]

    for ext_config in ext_modules:
        config.add_extension(**ext_config)

    config.add_subpackage('test')
    config.add_subpackage('app')

    return config


if __name__ == "__main__":
    from numpy.distutils.core import setup
    setup(configuration=configuration)



# -*- coding: utf-8 -*-
"""
Functions to generating graphs related to SAS.
"""

__authors__ = ["Jerome Kieffer"]
__license__ = "MIT"
__copyright__ = "2020, ESRF"
__date__ = "14/05/2020"

import logging

logger = logging.getLogger(__name__)
import numpy
from matplotlib.pyplot import subplots


def scatter_plot(
    data,
    guinier=None,
    ift=None,
    filename=None,
    img_format="svg",
    unit="nm",
    title="Scattering curve",
    ax=None,
    labelsize=None,
    fontsize=None,
):
    """
    Generate a scattering plot I = f(q) in semi_log_y.

    :param data: data read from an ASCII file, 3 column (q,I,err)
    :param filename: name of the file where the cuve should be saved
    :param img_format: image format
    :param unit: Unit name for Rg and 1/q
    :param guinier: output of autoRg
    :param ift: converged instance of BIFT (output of auto_bift)
    :param ax: subplot where the plot shall go in
    :return: the matplotlib figure
    """
    label_exp = "Experimental data"
    label_guinier = "Guinier region"
    label_ift = "BIFT extraplolated"
    exp_color = "blue"
    err_color = "lightblue"
    guinier_color = "limegreen"
    ift_color = "crimson"
    assert data.ndim == 2
    assert data.shape[1] >= 2
    q = data.T[0]
    I = data.T[1]
    try:
        err = data.T[2]
    except:
        err = None
    if ax:
        fig = ax.figure
    else:
        fig, ax = subplots()

    # Extend q to zero
    delta_q = (q[-1] - q[0]) / (len(q) - 1)
    extra_q = int(q[0] / delta_q)
    first = q[0] - extra_q * delta_q
    q_ext = numpy.linspace(first, q[-1], extra_q + len(q))

    if guinier is None:
        if ift is not None:
            # best = ift.calc_stats()[0]
            I0 = guinier.I0
            rg = guinier.rg
            first_point = ift.high_start
            last_point = ift.high.stop
        else:
            rg = I0 = first_point = last_point = None
    else:
        I0 = guinier.I0
        rg = guinier.Rg
        first_point = guinier.start_point
        last_point = guinier.end_point

    if (rg is None) and (ift is None):
        if err is not None:
            ax.errorbar(
                q,
                I,
                err,
                label=label_exp,
                capsize=0,
                color=exp_color,
                ecolor=err_color,
            )
        else:
            ax.plot(q, I, label=label_exp, color="blue")
    else:
        q_guinier = q[first_point:last_point]
        I_guinier = I0 * numpy.exp(-((q_guinier * rg) ** 2) / 3)
        if err is not None:
            ax.errorbar(
                q,
                I,
                err,
                label=label_exp,
                capsize=0,
                color=exp_color,
                ecolor=err_color,
                alpha=0.5,
            )
        else:
            ax.plot(q, I, label=label_exp, color=exp_color, alpha=0.5)
        label_guinier += ": $R_g=$%.2f %s, $I_0=$%.2f" % (rg, unit, I0)
        ax.plot(
            q_guinier,
            I_guinier,
            label=label_guinier,
            color=guinier_color,
            linewidth=5,
        )

    if ift:
        from ._bift import BIFT, StatsResult

        if isinstance(ift, BIFT):
            stats = ift.calc_stats()
        elif isinstance(ift, StatsResult):
            stats = ift
        else:
            raise TypeError("ift is expected to be a BIFT object")

        r = stats.radius
        T = numpy.outer(q_ext, r / numpy.pi)
        T = (4 * numpy.pi * (r[-1] - r[0]) / (len(r) - 1)) * numpy.sinc(T)
        p = stats.density_avg
        label_ift += ": $D_{max}=$%.2f %s,\n    $R_g=$%.2f %s, $I_0=$%.2f" % (
            stats.Dmax_avg,
            unit,
            stats.Rg_avg,
            unit,
            stats.I0_avg,
        )
        ax.plot(q_ext, T.dot(p), label=label_ift, color=ift_color)

    ax.set_ylabel("$I(q)$ (log scale)", fontsize=fontsize)
    ax.set_xlabel("$q$ (%s$^{-1}$)" % unit, fontsize=fontsize)
    ax.set_title(title)
    ax.set_yscale("log")
    #     ax.set_ylim(ymin=I.min() * 10, top=I.max() * 1.1)

    # Re-order labels ...
    crv, lab = ax.get_legend_handles_labels()
    ordered_lab = []
    ordered_crv = []
    for l in [label_exp, label_guinier, label_ift]:
        try:
            idx = lab.index(l)
        except:
            continue
        ordered_lab.append(lab[idx])
        ordered_crv.append(crv[idx])
    ax.legend(ordered_crv, ordered_lab, loc=3)
    ax.tick_params(axis="x", labelsize=labelsize)
    ax.tick_params(axis="y", labelsize=labelsize)
    if filename:
        if img_format:
            fig.savefig(filename, format=img_format)
        else:
            fig.savefig(filename)
    return fig


def kratky_plot(
    data,
    guinier,
    filename=None,
    img_format="svg",
    unit="nm",
    title="Dimensionless Kratky plot",
    ax=None,
    labelsize=None,
    fontsize=None,
):
    """
    Generate a Kratky plot q²Rg²I/I₀ = f(q·Rg)

    :param data: data read from an ASCII file, 3 column (q,I,err)
    :param guinier: output of autoRg
    :param filename: name of the file where the cuve should be saved
    :param img_format: image format
    :param unit: Unit name for Rg and 1/q
    :param ax: subplot where the plot shall go in
    :return: the matplotlib figure
    """
    label = "Experimental data"
    assert data.ndim == 2
    assert data.shape[1] >= 2
    q = data.T[0]
    I = data.T[1]
    try:
        err = data.T[2]
    except:
        err = None
    if ax:
        fig = ax.figure
    else:
        fig, ax = subplots()
    Rg = guinier.Rg
    I0 = guinier.I0

    xdata = q * Rg
    ydata = xdata * xdata * I / I0
    if err is not None:
        dy = xdata * xdata * err / I0
        dplot = ax.errorbar(
            xdata,
            ydata,
            dy,
            label=label,
            capsize=0,
            color="blue",
            ecolor="lightblue",
        )
    else:
        dplot = ax.plot(xdata, ydata, label=label, color="blue")
    ax.set_ylabel("$(qR_{g})^2 I/I_{0}$", fontsize=fontsize)
    ax.set_xlabel("$qR_{g}$", fontsize=fontsize)
    ax.legend(loc=1)

    ax.hlines(
        3.0 * numpy.exp(-1),
        xmin=-0.05,
        xmax=max(xdata),
        color="0.75",
        linewidth=1.0,
    )
    ax.vlines(
        numpy.sqrt(3.0),
        ymin=-0.01,
        ymax=max(ydata),
        color="0.75",
        linewidth=1.0,
    )
    ax.set_xlim(left=-0.05, right=8.5)
    ax.set_ylim(bottom=-0.01, top=(min(3.5, max(ydata))))
    ax.set_title(title)
    #     ax.legend([dplot[0]], [dplot[0].get_label()], loc=0)
    ax.legend(loc=0)
    ax.tick_params(axis="x", labelsize=labelsize)
    ax.tick_params(axis="y", labelsize=labelsize)

    if filename:
        if img_format:
            fig.savefig(filename, format=img_format)
        else:
            fig.savefig(filename)
    return fig


def guinier_plot(
    data,
    guinier,
    filename=None,
    img_format="png",
    unit="nm",
    ax=None,
    labelsize=None,
    fontsize=None,
):
    """
    Generate a guinier plot: ln(I) = f(q²)

    :param data: data read from an ASCII file, 3 column (q,I,err)
    :param guinier: A RG_RESULT object from AutoRg
    :param  filename: name of the file where the cuve should be saved
    :param img_format: image format
    :param: ax: subplot where to plot in
    :return: the matplotlib figure
    """
    assert data.ndim == 2
    assert data.shape[1] >= 2

    q, I, err = data.T[:3]

    mask = (I > 0) & numpy.isfinite(I) & (q > 0) & numpy.isfinite(q)
    if err is not None:
        mask &= (err > 0.0) & numpy.isfinite(err)
    mask = mask.astype(bool)
    Rg = guinier.Rg
    I0 = guinier.I0
    first_point = guinier.start_point
    last_point = guinier.end_point
    intercept = numpy.log(I0)
    slope = -Rg * Rg / 3.0
    end = numpy.where(q > 1.5 / Rg)[0][0]
    mask[end:] = False

    q2 = q[mask] ** 2
    logI = numpy.log(I[mask])

    if ax:
        fig = ax.figure
    else:
        fig, ax = subplots(figsize=(12, 10))
    if err is not None:
        dlogI = err[mask] / logI
        ax.errorbar(
            q2,
            logI,
            dlogI,
            label="Experimental curve",
            capsize=0,
            color="blue",
            ecolor="lightblue",
            alpha=0.5,
        )
    else:
        ax.plot(
            q2[mask],
            logI[mask],
            label="Experimental curve",
            color="blue",
            alpha=0.5,
        )
    # ax.plot(q2[first_point:last_point], logI[first_point:last_point], marker='D', markersize=5, label="guinier region")
    xmin = q[first_point] ** 2
    xmax = q[last_point] ** 2
    ymax = numpy.log(I[first_point])
    ymin = numpy.log(I[last_point])
    dy = (ymax - ymin) / 2.0
    ax.vlines(xmin, ymin=ymin, ymax=ymax + dy, color="0.75", linewidth=1.0)
    ax.vlines(
        xmax, ymin=ymin - dy, ymax=ymin + dy, color="0.75", linewidth=1.0
    )
    ax.annotate(
        "$(qR_{g})_{min}$=%.1f" % (Rg * q[first_point]),
        (xmin, ymax + dy),
        xytext=None,
        xycoords="data",
        textcoords="data",
    )
    ax.annotate(
        "$(qR_{g})_{max}$=%.1f" % (Rg * q[last_point]),
        (xmax, ymin + dy),
        xytext=None,
        xycoords="data",
        textcoords="data",
    )
    ax.annotate(
        "Guinier region",
        (xmin, ymin - dy),
        xytext=None,
        xycoords="data",
        textcoords="data",
    )
    ax.plot(
        q2[:end],
        intercept + slope * q2[:end],
        label="ln[$I(q)$] = %.2f %.2f * $q^2$" % (intercept, slope),
        color="crimson",
    )
    ax.set_ylabel("ln[$I(q)$]", fontsize=fontsize)
    ax.set_xlabel("$q^2$ (%s$^{-2}$)" % unit, fontsize=fontsize)
    ax.set_title("Guinier plot: $R_{g}=$%.2f %s $I_{0}=$%.2f" % (Rg, unit, I0))
    ax.legend()
    ax.tick_params(axis="x", labelsize=labelsize)
    ax.tick_params(axis="y", labelsize=labelsize)

    if filename:
        if img_format:
            fig.savefig(filename, format=img_format)
        else:
            fig.savefig(filename)
    return fig


def density_plot(
    ift,
    filename=None,
    img_format="png",
    unit="nm",
    ax=None,
    labelsize=None,
    fontsize=None,
):
    """
    Generate a density plot p(r)

    @param ift: An IFT result comming out of BIFT
    @param  filename: name of the file where the cuve should be saved
    @param img_format: image image format
    @param ax: subplotib where to plot in
    @return: the matplotlib figure
    """
    if ax:
        fig = ax.figure
    else:
        fig, ax = subplots(figsize=(12, 10))

    from ._bift import BIFT, StatsResult

    if isinstance(ift, BIFT):
        stats = ift.calc_stats()
    elif isinstance(ift, StatsResult):
        stats = ift
    else:
        raise TypeError("ift is expected to be a BIFT object")

    ax.errorbar(
        ift.radius,
        ift.density_avg,
        ift.density_std,
        label="BIFT: χ$_{r}^{2}=$%.2f\n $D_{max}=$%.2f %s\n $R_{g}=$%.2f %s\n $I_{0}=$%.2f"
        % (
            stats.chi2r_avg,
            stats.Dmax_avg,
            unit,
            stats.Rg_avg,
            unit,
            stats.I0_avg,
        ),
        capsize=0,
        color="blue",
        ecolor="lightblue",
    )
    ax.set_ylabel("$p(r)$", fontsize=fontsize)
    ax.set_xlabel("$r$ (%s)" % unit, fontsize=fontsize)
    ax.set_title("Pair distribution function")
    ax.legend()
    ax.tick_params(axis="x", labelsize=labelsize)
    ax.tick_params(axis="y", labelsize=labelsize)

    if filename:
        if img_format:
            fig.savefig(filename, format=img_format)
        else:
            fig.savefig(filename)
    return fig


def plot_all(
    data,
    filename=None,
    img_format=None,
    unit="nm",
    labelsize=None,
    fontsize=None,
):
    from . import bift, autorg

    try:
        guinier = autorg.autoRg(data)
    except autorg.InsufficientDataError:
        raise
    logger.debug(guinier)
    try:
        bo = bift.auto_bift(data, npt=100, scan_size=11, Dmax_over_Rg=3)
    except (
        autorg.InsufficientDataError,
        autorg.NoGuinierRegionError,
        ValueError,
    ):
        raise
    else:
        ift = bo.calc_stats()
    logger.debug(ift)
    fig, ax = subplots(2, 2, figsize=(12, 10))
    scatter_plot(
        data,
        guinier=guinier,
        ift=ift,
        ax=ax[0, 0],
        unit=unit,
        labelsize=labelsize,
        fontsize=fontsize,
    )
    guinier_plot(
        data,
        guinier,
        filename=None,
        img_format=None,
        unit=unit,
        ax=ax[0, 1],
        labelsize=labelsize,
        fontsize=fontsize,
    )
    kratky_plot(
        data,
        guinier,
        filename=None,
        img_format=None,
        unit=unit,
        ax=ax[1, 0],
        labelsize=labelsize,
        fontsize=fontsize,
    )
    density_plot(
        ift,
        filename=None,
        img_format=None,
        unit=unit,
        ax=ax[1, 1],
        labelsize=labelsize,
        fontsize=fontsize,
    )
    if filename is not None:
        if img_format:
            fig.savefig(filename, format=img_format)
        else:
            fig.savefig(filename)
    return fig



#!/usr/bin/python3
# coding: utf-8

__author__ = "Jérôme Kieffer"
__license__ = "MIT"
__copyright__ = "2015, ESRF"
__date__ = "20/04/2020"

import os
import logging
import freesas
from freesas.cormap import gof

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("cormap")
import numpy
from itertools import combinations
from collections import namedtuple
from freesas.sasio import load_scattering_data
from freesas.sas_argparser import SASParser

datum = namedtuple("datum", ["index", "filename", "data"])

import platform

operatingSystem = platform.system()
if operatingSystem == "Windows":
    import glob


def parse():
    """Parse input and return list of files.
    :return: list of input files
    """
    description = "Measure pair-wise similarity of spectra "
    epilog = """cormapy is an open-source implementation of
    the cormap algorithm in datcmp (from ATSAS).
    It does not scale the data and assume they are already scaled
    """
    parser = SASParser(prog="cormapy", description=description, epilog=epilog)
    parser.add_file_argument(help_text="dat files to compare")

    args = parser.parse_args()

    if args.verbose:
        logging.root.setLevel(logging.DEBUG)
    files = [i for i in args.file if os.path.exists(i)]
    if operatingSystem == "Windows" and files == []:
        files = glob.glob(args.file[0])
        files.sort()
    input_len = len(files)
    logger.debug("%s input files" % input_len)
    return files


def compare(lstfiles):
    res = [
        "Pair-wise Correlation Map",
        "" "                                C       Pr(>C)",
    ]
    data = []
    for i, f in enumerate(lstfiles):
        try:
            ary = load_scattering_data(f)
        except ValueError as e:
            print(e)
        if ary.ndim > 1 and ary.shape[1] > 1:
            ary = ary[:, 1]
        d = datum(i + 1, f, ary)
        data.append(d)
    for a, b in combinations(data, 2):
        r = gof(a.data, b.data)
        res.append(
            "%6i vs. %6i          %6i     %8.6f" % (a.index, b.index, r.c, r.P)
        )
    res.append("")
    for a in data:
        res.append(
            "%6i         %8f + %8f * %s" % (a.index, 0.0, 1.0, a.filename)
        )
    res.append("")
    print(os.linesep.join(res))
    return res


def main():
    """main entry point"""
    f = parse()
    if f:
        compare(f)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Generalized arg parser for freeSAS apps to ensure unified command line API.
"""

__author__ = "Martha Brennich"
__license__ = "MIT"
__copyright__ = "2020, ESRF"
__date__ = "09/08/2020"

import argparse
from pathlib import Path
from freesas import dated_version as freesas_version


def parse_unit(unit_input: str) -> str:
    """
    Parser for sloppy acceptance of unit flags.
    Current rules:
    "A" ➔ "Å"
    :param unit_input: unit flag as provided by the user
    :return: cast of user input to known flag if sloppy rule defined,
             else user input.
    """
    if unit_input == "A":  # pylint: disable=R1705
        return "Å"
    else:
        return unit_input


class SASParser:

    """
    Wrapper class for argparse ArgumentParser that provides predefined argument.
    """

    usage = ""

    def __init__(self, prog: str, description: str, epilog: str, **kwargs):
        """
        Create parser argparse ArgumentParser
        - standardized usage text
        - standardized verion text
        - verbose and version args added by default

        :param prog:         name of the executable
        :param description:  description param of argparse ArgumentParser
        :param epilog:       epilog param of argparse ArgumentParser
        :param kwargs:       additional kwargs for argparse ArgumentParser
        """

        self.usage = "%s [OPTIONS] FILES " % (prog)
        version = "%s version %s from %s" % (
            prog,
            freesas_version.version,
            freesas_version.date,
        )

        self.parser = argparse.ArgumentParser(
            usage=self.usage, description=description, epilog=epilog, **kwargs
        )
        self.add_argument(
            "-v",
            "--verbose",
            default=0,
            help="switch to verbose mode",
            action="count",
        )
        self.add_argument("-V", "--version", action="version", version=version)

    def parse_args(self, *args, **kwargs):
        """ Wrapper for argparse parse_args() """
        return self.parser.parse_args(*args, **kwargs)

    def add_argument(self, *args, **kwargs):
        """ Wrapper for argparse add_argument() """
        self.parser.add_argument(*args, **kwargs)

    def add_file_argument(self, help_text: str):
        """
        Add positional file argument.

        :param help_text: specific help text to be displayed
        """
        self.add_argument("file", metavar="FILE", nargs="+", help=help_text)

    def add_q_unit_argument(self):
        """
        Add default argument for selecting length unit of input data
        between Å and nm. nm is default.
        """
        self.add_argument(
            "-u",
            "--unit",
            action="store",
            choices=["nm", "Å", "A"],
            help="Unit for q: inverse nm or Ångstrom?",
            default="nm",
            type=parse_unit,
        )

    def add_output_filename_argument(self):
        """ Add default argument for specifying output format. """
        self.add_argument(
            "-o",
            "--output",
            action="store",
            help="Output filename",
            default=None,
            type=Path,
        )

    def add_output_data_format(self, *formats: str, default: str = None):
        """ Add default argument for specifying output format. """
        help_string = "Output format: " + ", ".join(formats)
        self.add_argument(
            "-f",
            "--format",
            action="store",
            help=help_string,
            default=default,
            type=str,
        )


class GuinierParser:
    """
    Wrapper class for argparse ArgumentParser that provides predefined
    arguments for auto_rg like programs.
    """

    usage = ""

    def __init__(self, prog: str, description: str, epilog: str, **kwargs):
        """
        Create parser argparse ArgumentParser with argument
        - standardized usage text
        - standardized version text
        - verbose and version args added by default

        :param prog:         name of the executable
        :param description:  description param of argparse ArgumentParser
        :param epilog:       epilog param of argparse ArgumentParser
        :param kwargs:       additional kwargs for argparse ArgumentParser
        """

        file_help_text = "dat files of the scattering curves"
        self.parser = SASParser(
            prog=prog, description=description, epilog=epilog, **kwargs
        )
        self.parser.add_file_argument(help_text=file_help_text)
        self.parser.add_output_filename_argument()
        self.parser.add_output_data_format(
            "native", "csv", "ssf", default="native"
        )
        self.parser.add_q_unit_argument()
        self.usage = self.parser.usage

    def parse_args(self, *args, **kwargs):
        """ Wrapper for SASParser parse_args() """
        return self.parser.parse_args(*args, **kwargs)

    def add_argument(self, *args, **kwargs):
        """ Wrapper for SASParser add_argument() """
        self.parser.add_argument(*args, **kwargs)



# coding: utf-8
#
#    Project: Free SAS tools
#             https://github.com/kif/freesas
#
#    Copyright (C) 2015-2020 European Synchrotron Radiation Facility, Grenoble, France
#
#    Principal author:       Jérôme Kieffer (Jerome.Kieffer@ESRF.eu)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Bunch of useful decorators"""

__authors__ = ["Jerome Kieffer", "H. Payno", "P. Knobel", "V. Valls"]
__contact__ = "Jerome.Kieffer@ESRF.eu"
__license__ = "MIT"
__copyright__ = "European Synchrotron Radiation Facility, Grenoble, France"
__date__ = "27/04/2020"
__status__ = "development"
__docformat__ = 'restructuredtext'

import sys
import time
import logging

timelog = logging.getLogger("freesas.timeit")


def timeit(func):

    def wrapper(*arg, **kw):
        '''This is the docstring of timeit:
        a decorator that logs the execution time'''
        t1 = time.perf_counter()
        res = func(*arg, **kw)
        t2 = time.perf_counter()
        name = func.func_name if sys.version_info[0] < 3 else func.__name__
        timelog.warning("%s took %.3fs", name, t2 - t1)
        return res

    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__
    return wrapper



#!/usr/bin/python3
# coding: utf-8
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2017  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__author__ = "Jérôme Kieffer"
__license__ = "MIT"
__copyright__ = "2017, ESRF"
__date__ = "13/10/2020"

import sys
import logging
import platform
import traceback
from os import linesep
from pathlib import Path
from numpy import float32
from freesas import bift
from freesas.sasio import (
    load_scattering_data,
    convert_inverse_angstrom_to_nanometer,
)
from freesas.sas_argparser import SASParser

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger("bift")


def parse():
    """Parse input and return list of files.
    :return: list of input files
    """
    description = (
        "Calculate the density as function of distance p(r)"
        " curve from an I(q) scattering curve"
    )
    epilog = """free_bift is a Python implementation of the Bayesian Inverse Fourier Transform

    This code is the implementation of
    Steen Hansen J. Appl. Cryst. (2000). 33, 1415-1421

    Based on the BIFT from Jesse Hopkins, available at:
    https://sourceforge.net/p/bioxtasraw/git/ci/master/tree/bioxtasraw/BIFT.py

    It aims at being a drop in replacement for datgnom of the ATSAS suite.

    """
    parser = SASParser(
        prog="free_bift", description=description, epilog=epilog
    )
    parser.add_file_argument(help_text="I(q) files to convert into p(r)")
    parser.add_output_filename_argument()
    parser.add_q_unit_argument()
    parser.add_argument(
        "-n",
        "--npt",
        default=100,
        type=int,
        help="number of points in p(r) curve",
    )
    parser.add_argument(
        "-s",
        "--scan",
        default=27,
        type=int,
        help="Initial alpha-scan size to guess the start parameter",
    )
    parser.add_argument(
        "-m",
        "--mc",
        default=100,
        type=int,
        help="Number of Monte-Carlo samples in post-refinement",
    )
    parser.add_argument(
        "-t",
        "--threshold",
        default=2.0,
        type=float,
        help="Sample at average ± threshold*sigma in MC",
    )

    args = parser.parse_args()

    if args.verbose:
        logging.root.setLevel(logging.DEBUG)
    files = [Path(i) for i in args.file if Path(i).exists()]
    if platform.system() == "Windows" and files == []:
        files = list(Path.cwd().glob(args.file[0]))
        files.sort()
    input_len = len(files)
    logger.debug("%s input files" % input_len)
    return files, args


def main():
    if platform.system() == "Windows":
        sys.stdout = open(1, "w", encoding="utf-16", closefd=False)
    list_files, args = parse()
    for afile in list_files:
        try:
            data = load_scattering_data(afile)
        except:
            logger.error("Unable to parse file %s", afile)
        else:
            if args.unit == "Å":
                data = convert_inverse_angstrom_to_nanometer(data)
            try:
                bo = bift.auto_bift(data, npt=args.npt, scan_size=args.scan)
            except Exception as err:
                print("%s: %s %s" % (afile, err.__class__.__name__, err))
                if logging.root.level < logging.WARNING:
                    traceback.print_exc(file=sys.stdout)
            else:
                try:
                    stats = bo.monte_carlo_sampling(
                        args.mc, args.threshold, npt=args.npt
                    )
                except RuntimeError as err:
                    print("%s: %s %s" % (afile, err.__class__.__name__, err))
                    if logging.root.level < logging.WARNING:
                        traceback.print_exc(file=sys.stdout)
                else:
                    dest = afile.stem + ".out"
                    print(stats.save(dest, source=afile))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
#    Project: freesas
#             https://github.com/kif/freesas
#
#    Copyright (C) 2017  European Synchrotron Radiation Facility, Grenoble, France
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

__authors__ = ["Martha Brennich"]
__license__ = "MIT"
__date__ = "25/07/2020"

import unittest
import logging
from sys import version_info
from numpy import array, allclose
from ..sasio import parse_ascii_data, load_scattering_data, \
                    convert_inverse_angstrom_to_nanometer
if version_info.minor > 6:
    from unittest.mock import mock_open, patch
else:
    from .mock_open_38 import mock_open
    from unittest.mock import patch
logger = logging.getLogger(__name__)

class TestSasIO(unittest.TestCase):

    def test_parse_3_ok(self):
        """
        Test for successful parsing of file with some invalid lines
        """
        file_content = ["Test data for",
                        "file parsing",
                        "1 1 1",
                        "2 a 2",
                        "3 3 3",
                        "some stuff at the end",
                        ]
        expected_result = array([[1.0, 1.0, 1.0], [3.0, 3.0, 3.0]])
        data = parse_ascii_data(file_content, number_of_columns=3)
        self.assertTrue(allclose(data, expected_result, 1e-7),
                        msg="3 column parse returns expected result")

    def test_parse_no_data(self):
        """
        Test that an empty input list raises a ValueError
        """
        file_content = []
        with self.assertRaises(ValueError, msg="Empty list cannot be parsed"):
            parse_ascii_data(file_content, number_of_columns=3)

    def test_parse_no_valid_data(self):
        """
        Test that an input list with no valid data raises a ValueError
        """
        file_content = ["a a a", "2 4", "3 4 5 6", "# 3 4 6"]
        with self.assertRaises(ValueError,
                               msg="File with no float float float data"
                                   " cannot be parsed"):
            parse_ascii_data(file_content, number_of_columns=3)

    def test_load_clean_data(self):
        """
        Test that clean float float float data is loaded correctly.
        """
        file_content = ["# Test data for"
                        "# file parsing",
                        "1 1 1",
                        "2.0 2.0 1.0",
                        "3 3 3",
                        "#REMARK some stuff at the end",
                        ]
        expected_result = array([[1.0, 1.0, 1.0],
                                 [2.0, 2.0, 1.0],
                                 [3.0, 3.0, 3.0]])
        file_data = "\n".join(file_content)
        mocked_open = mock_open(read_data=file_data)
        with patch('builtins.open', mocked_open):
            with patch('numpy.DataSource.open', mocked_open):
                data = load_scattering_data("test")
        self.assertTrue(allclose(data, expected_result, 1e-7),
                        msg="Sunny data loaded correctly")

    def test_load_data_with_unescaped_header(self):
        """
        Test that an unescaped header does not hinder loading.
        """
        file_content = ["Test data for",
                        "file parsing",
                        "1 1 1",
                        "2.0 2.0 1.0",
                        "3 3 3",
                        ]
        expected_result = array([[1.0, 1.0, 1.0],
                                 [2.0, 2.0, 1.0],
                                 [3.0, 3.0, 3.0]])
        file_data = "\n".join(file_content)
        mocked_open = mock_open(read_data=file_data)
        with patch('builtins.open', mocked_open), \
             patch('numpy.DataSource.open', mocked_open):
            data = load_scattering_data("test")
        self.assertTrue(allclose(data, expected_result, 1e-7),
                        msg="Sunny data loaded correctly")

    def test_load_data_with_unescaped_footer(self):
        """
        Test that an unescaped footer does not hinder loading.
        """
        file_content = [
                        "1 1 1",
                        "2.0 2.0 1.0",
                        "3 3 3",
                        "REMARK some stuff at the end"
                        ]
        expected_result = array([[1.0, 1.0, 1.0],
                                 [2.0, 2.0, 1.0],
                                 [3.0, 3.0, 3.0]])
        file_data = "\n".join(file_content)
        mocked_open = mock_open(read_data=file_data)
        with patch('builtins.open', mocked_open), \
             patch('numpy.DataSource.open', mocked_open):
            data = load_scattering_data("test")
        self.assertTrue(allclose(data, expected_result, 1e-7),
                        msg="Sunny data loaded correctly")


    def test_load_invalid_data(self):
        """
        Test that invalid data raises a ValueError.
        """
        file_content = ["a a a", "2 4", "3 4 5 6", "# 3 4 6"]
        file_data = "\n".join(file_content)
        mocked_open = mock_open(read_data=file_data)

        with patch('builtins.open', mocked_open), \
             patch('numpy.DataSource.open', mocked_open), \
             self.assertRaises(ValueError,
                               msg="File with no float float float "
                                   "data cannot be loaded"):
            load_scattering_data("test")

    def test_convert_inverse_angstrom_to_nanometer(self):
        """
        Test conversion of data with q in 1/Å to 1/nm
        """
        input_data = array([[1.0, 1.0, 1.0],
                            [2.0, 2.0, 1.0],
                            [3.0, 3.0, 3.0]])
        expected_result = array([[10, 1.0, 1.0],
                                 [20, 2.0, 1.0],
                                 [30, 3.0, 3.0]])
        result = convert_inverse_angstrom_to_nanometer(input_data)
        self.assertTrue(allclose(result, expected_result, 1e-7),
                        msg="Converted to 1/nm from 1/Å")

    def test_unit_conversion_creates_new_array(self):
        """
        Test conversion of data does not change original data
        """
        input_data = array([[1.0, 1.0, 1.0],
                            [2.0, 2.0, 1.0],
                            [3.0, 3.0, 3.0]])
        expected_data = array([[1.0, 1.0, 1.0],
                               [2.0, 2.0, 1.0],
                               [3.0, 3.0, 3.0]])
        _ = convert_inverse_angstrom_to_nanometer(input_data)
        self.assertTrue(allclose(input_data, expected_data, 1e-7),
                        msg="Conversion function does not change its input")


def suite():
    test_suite = unittest.TestSuite()
    test_suite.addTest(TestSasIO("test_parse_3_ok"))
    test_suite.addTest(TestSasIO("test_parse_no_data"))
    test_suite.addTest(TestSasIO("test_parse_no_valid_data"))
    test_suite.addTest(TestSasIO("test_load_clean_data"))
    test_suite.addTest(TestSasIO("test_load_data_with_unescaped_header"))
    test_suite.addTest(TestSasIO("test_load_data_with_unescaped_footer"))
    test_suite.addTest(TestSasIO("test_load_invalid_data"))
    test_suite.addTest(TestSasIO("test_convert_inverse_angstrom_to_nanometer"))
    test_suite.addTest(TestSasIO("test_unit_conversion_creates_new_array"))
    return test_suite


if __name__ == '__main__':
    runner = unittest.TextTestRunner()
    runner.run(suite())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Bootstrap helps you to test scripts without installing them
by patching your PYTHONPATH on the fly

example: ./bootstrap.py ipython
"""

__authors__ = ["Frédéric-Emmanuel Picca", "Jérôme Kieffer"]
__contact__ = "jerome.kieffer@esrf.eu"
__license__ = "MIT"
__date__ = "09/07/2020"

import sys
import os
import distutils.util
import subprocess
import logging
import collections
from argparse import ArgumentParser

logging.basicConfig()
logger = logging.getLogger("bootstrap")


def is_debug_python():
    """Returns true if the Python interpreter is in debug mode."""
    try:
        import sysconfig
    except ImportError:  # pragma nocover
        # Python < 2.7
        import distutils.sysconfig as sysconfig

    if sysconfig.get_config_var("Py_DEBUG"):
        return True

    return hasattr(sys, "gettotalrefcount")


def _distutils_dir_name(dname="lib"):
    """
    Returns the name of a distutils build directory
    """
    platform = distutils.util.get_platform()
    architecture = "%s.%s-%i.%i" % (dname, platform,
                                    sys.version_info[0], sys.version_info[1])
    if is_debug_python():
        architecture += "-pydebug"
    return architecture


def _distutils_scripts_name():
    """Return the name of the distrutils scripts sirectory"""
    f = "scripts-{version[0]}.{version[1]}"
    return f.format(version=sys.version_info)


def _get_available_scripts(path):
    res = []
    try:
        res = " ".join([s.rstrip('.py') for s in os.listdir(path)])
    except OSError:
        res = ["no script available, did you ran "
               "'python setup.py build' before bootstrapping ?"]
    return res


if sys.version_info[0] >= 3:  # Python3

    def execfile(fullpath, globals=None, locals=None):
        "Python3 implementation for execfile"
        with open(fullpath) as f:
            try:
                data = f.read()
            except UnicodeDecodeError:
                raise SyntaxError("Not a Python script")
            code = compile(data, fullpath, 'exec')
            exec(code, globals, locals)


def run_file(filename, argv):
    """
    Execute a script trying first to use execfile, then a subprocess

    :param str filename: Script to execute
    :param list[str] argv: Arguments passed to the filename
    """
    full_args = [filename]
    full_args.extend(argv)

    try:
        logger.info("Execute target using exec")
        # execfile is considered as a local call.
        # Providing globals() as locals will force to feed the file into
        # globals() (for examples imports).
        # Without this any function call from the executed file loses imports
        try:
            old_argv = sys.argv
            sys.argv = full_args
            logger.info("Patch the sys.argv: %s", sys.argv)
            logger.info("Executing %s.main()", filename)
            print("########### EXECFILE ###########")
            module_globals = globals().copy()
            module_globals['__file__'] = filename
            execfile(filename, module_globals, module_globals)
        finally:
            sys.argv = old_argv
    except SyntaxError as error:
        logger.error(error)
        logger.info("Execute target using subprocess")
        env = os.environ.copy()
        env.update({"PYTHONPATH": LIBPATH + os.pathsep + os.environ.get("PYTHONPATH", ""),
                    "PATH": os.environ.get("PATH", "")})
        print("########### SUBPROCESS ###########")
        run = subprocess.Popen(full_args, shell=False, env=env)
        run.wait()


def run_entry_point(entry_point, argv):
    """
    Execute an entry_point using the current python context
    (http://setuptools.readthedocs.io/en/latest/setuptools.html#automatic-script-creation)

    :param str entry_point: A string identifying a function from a module
        (NAME = PACKAGE.MODULE:FUNCTION [EXTRA])
    """
    import importlib
    elements = entry_point.split("=")
    target_name = elements[0].strip()
    elements = elements[1].split(":")
    module_name = elements[0].strip()
    # Take care of entry_point optional "extra" requirements declaration
    function_name = elements[1].split()[0].strip()

    logger.info("Execute target %s (function %s from module %s) using importlib", target_name, function_name, module_name)
    full_args = [target_name]
    full_args.extend(argv)
    try:
        old_argv = sys.argv
        sys.argv = full_args
        print("########### IMPORTLIB ###########")
        module = importlib.import_module(module_name)
        if hasattr(module, function_name):
            func = getattr(module, function_name)
            func()
        else:
            logger.info("Function %s not found", function_name)
    finally:
        sys.argv = old_argv


def find_executable(target):
    """Find a filename from a script name.

    - Check the script name as file path,
    - Then checks if the name is a target of the setup.py
    - Then search the script from the PATH environment variable.

    :param str target: Name of the script
    :returns: Returns a tuple: kind, name.
    """
    if os.path.isfile(target):
        return ("path", os.path.abspath(target))

    # search the file from setup.py
    import setup
    config = setup.get_project_configuration(dry_run=True)
    # scripts from project configuration
    if "scripts" in config:
        for script_name in config["scripts"]:
            if os.path.basename(script_name) == target:
                return ("path", os.path.abspath(script_name))
    # entry-points from project configuration
    if "entry_points" in config:
        for kind in config["entry_points"]:
            for entry_point in config["entry_points"][kind]:
                elements = entry_point.split("=")
                name = elements[0].strip()
                if name == target:
                    return ("entry_point", entry_point)

    # search the file from env PATH
    for dirname in os.environ.get("PATH", "").split(os.pathsep):
        path = os.path.join(dirname, target)
        if os.path.isfile(path):
            return ("path", path)

    return None, None


def main(argv):
    parser = ArgumentParser(prog="bootstrap", usage="./bootstrap.py <script>",
                            description=__doc__)
    parser.add_argument("script", nargs="*")
    parser.add_argument("-m", help="run library module as a script (terminates option list)")

    Options = collections.namedtuple("Options", ["script", "module"])
    if len(argv) == 1:
        options = Options(script=None, module=None)
    else:
        if argv[1] in ["-h", "--help"]:
            parser.print_help()
            return
        if argv[1] == "-m":
            if len(argv) < 3:
                parser.parse_args(argv[1:])
                return
            options = Options(script=None, module=argv[2:])
        else:
            options = Options(script=argv[1:], module=None)

    if options.script is not None:
        logger.info("Executing %s from source checkout", options.script)
        script = options.script[0]
        argv = options.script[1:]
        kind, target = find_executable(script)
        if kind == "path":
            run_file(target, argv)
        elif kind == "entry_point":
            run_entry_point(target, argv)
        else:
            logger.error("Script %s not found", options.script)
    elif options.module is not None:
        logging.info("Running module %s", options.module)
        import runpy
        module = options.module[0]
        try:
            old = sys.argv
            sys.argv = [None] + options.module[1:]
            runpy.run_module(module, run_name="__main__", alter_sys=True)
        finally:
            sys.argv = old
    else:
        logging.info("Running IPython by default")
        logger.info("Patch the sys.argv: %s", sys.argv)
        sys.path.insert(2, "")
        try:
            from IPython import embed
        except Exception as err:
            logger.error("Unable to execute iPython, using normal Python")
            logger.error(err)
            import code
            code.interact()
        else:
            embed()


if __name__ == "__main__":
    home = os.path.dirname(os.path.abspath(__file__))
    LIBPATH = os.path.join(home, 'build', _distutils_dir_name('lib'))
    cwd = os.getcwd()
    os.chdir(home)
    build = subprocess.Popen([sys.executable, "setup.py", "build"],
                         shell=False, cwd=os.path.dirname(os.path.abspath(__file__)))
    build_rc = build.wait()
    if not os.path.exists(LIBPATH):
        logger.warning("`lib` directory does not exist, trying common Python3 lib")
        LIBPATH = os.path.join(os.path.split(LIBPATH)[0], "lib")
    os.chdir(cwd)

    if build_rc == 0:
        logger.info("Build process ended.")
    else:
        logger.error("Build process ended with rc=%s", build_rc)
        sys.exit(-1)

    sys.path.insert(0, LIBPATH)
    logger.info("Patched sys.path with %s", LIBPATH)

    main(sys.argv)

